"""
Standalone chat templates extracted from Unsloth.
This file works without GPU dependencies.

Source: Extracted from unsloth/chat_templates.py
Total templates: 32

NOTE: These templates are used for LISTING available templates only.
The actual message formatting uses:
- CUDA: Unsloth's get_chat_template()
- CPU/MPS: Tokenizer's native template via safe_apply_chat_template()
"""

# Chat templates dictionary
CHAT_TEMPLATES = {}

# alpaca template
CHAT_TEMPLATES["alpaca"] = (
    """{{ bos_token }}{% if messages[0]['role'] == 'system' %}{{ messages[0]['content'] + '\n\n' }}{% set loop_messages = messages[1:] %}{% else %}{{ '{system_message}' + '\n\n' }}{% set loop_messages = messages %}{% endif %}{% for message in loop_messages %}{% if message['role'] == 'user' %}{{ '### Instruction:\n' + message['content'] + '\n\n' }}{% elif message['role'] == 'assistant' %}{{ '### Response:\n' + message['content'] + eos_token + '\n\n' }}{% else %}{{ raise_exception('Only user and assistant roles are supported!') }}{% endif %}{% endfor %}{% if add_generation_prompt %}{{ '### Response:\n' }}{% endif %}"""
)

# chatml template
CHAT_TEMPLATES["chatml"] = (
    """{% for message in messages %}{% if message['role'] == 'user' %}{{'<|im_start|>user\n' + message['content'] + '<|im_end|>\n'}}{% elif message['role'] == 'assistant' %}{{'<|im_start|>assistant\n' + message['content'] + '<|im_end|>\n' }}{% else %}{{ '<|im_start|>system\n' + message['content'] + '<|im_end|>\n' }}{% endif %}{% endfor %}{% if add_generation_prompt %}{{ '<|im_start|>assistant\n' }}{% endif %}"""
)

# gemma template
CHAT_TEMPLATES["gemma"] = (
    """{{ bos_token }}{% if messages[0]['role'] == 'system' %}{{'<start_of_turn>user\n' + messages[0]['content'] | trim + ' ' + messages[1]['content'] | trim + '<end_of_turn>\n'}}{% set messages = messages[2:] %}{% endif %}{% for message in messages %}{% if message['role'] == 'user' %}{{'<start_of_turn>user\n' + message['content'] | trim + '<end_of_turn>\n'}}{% elif message['role'] == 'assistant' %}{{'<start_of_turn>model\n' + message['content'] | trim + '<end_of_turn>\n' }}{% else %}{{ raise_exception('Only user and assistant roles are supported!') }}{% endif %}{% endfor %}{% if add_generation_prompt %}{{ '<start_of_turn>model\n' }}{% endif %}"""
)

# gemma-3 template
CHAT_TEMPLATES[
    "gemma-3"
] = """{{ bos_token }}
{%- if messages[0]['role'] == 'system' -%}
    {%- if messages[0]['content'] is string -%}
        {%- set first_user_prefix = messages[0]['content'] + '\n\n' -%}
    {%- else -%}
        {%- set first_user_prefix = messages[0]['content'][0]['text'] + '\n\n' -%}
    {%- endif -%}
    {%- set loop_messages = messages[1:] -%}
{%- else -%}
    {%- set first_user_prefix =  -%}
    {%- set loop_messages = messages -%}
{%- endif -%}
{%- for message in loop_messages -%}
    {%- if (message['role'] == 'user') != (loop.index0 % 2 == 0) -%}
        {{ raise_exception("""

# gemma-3n template
CHAT_TEMPLATES[
    "gemma-3n"
] = """{{ bos_token }}
{%- if messages[0]['role'] == 'system' -%}
    {%- if messages[0]['content'] is string -%}
        {%- set first_user_prefix = messages[0]['content'] + '\n\n' -%}
    {%- else -%}
        {%- set first_user_prefix = messages[0]['content'][0]['text'] + '\n\n' -%}
    {%- endif -%}
    {%- set loop_messages = messages[1:] -%}
{%- else -%}
    {%- set first_user_prefix =  -%}
    {%- set loop_messages = messages -%}
{%- endif -%}
{%- for message in loop_messages -%}
    {%- if (message['role'] == 'user') != (loop.index0 % 2 == 0) -%}
        {{ raise_exception("""

# gemma2 template
CHAT_TEMPLATES["gemma2"] = (
    """{{ bos_token }}{% if messages[0]['role'] == 'system' %}{{'<start_of_turn>user\n' + messages[0]['content'] | trim + ' ' + messages[1]['content'] | trim + '<end_of_turn>\n'}}{% set messages = messages[2:] %}{% endif %}{% for message in messages %}{% if message['role'] == 'user' %}{{'<start_of_turn>user\n' + message['content'] | trim + '<end_of_turn>\n'}}{% elif message['role'] == 'assistant' %}{{'<start_of_turn>model\n' + message['content'] | trim + '<end_of_turn>\n' }}{% else %}{{ raise_exception('Only user and assistant roles are supported!') }}{% endif %}{% endfor %}{% if add_generation_prompt %}{{ '<start_of_turn>model\n' }}{% endif %}"""
)

# gemma3 template
CHAT_TEMPLATES[
    "gemma3"
] = """{{ bos_token }}
{%- if messages[0]['role'] == 'system' -%}
    {%- if messages[0]['content'] is string -%}
        {%- set first_user_prefix = messages[0]['content'] + '\n\n' -%}
    {%- else -%}
        {%- set first_user_prefix = messages[0]['content'][0]['text'] + '\n\n' -%}
    {%- endif -%}
    {%- set loop_messages = messages[1:] -%}
{%- else -%}
    {%- set first_user_prefix =  -%}
    {%- set loop_messages = messages -%}
{%- endif -%}
{%- for message in loop_messages -%}
    {%- if (message['role'] == 'user') != (loop.index0 % 2 == 0) -%}
        {{ raise_exception("""

# gemma3n template
CHAT_TEMPLATES[
    "gemma3n"
] = """{{ bos_token }}
{%- if messages[0]['role'] == 'system' -%}
    {%- if messages[0]['content'] is string -%}
        {%- set first_user_prefix = messages[0]['content'] + '\n\n' -%}
    {%- else -%}
        {%- set first_user_prefix = messages[0]['content'][0]['text'] + '\n\n' -%}
    {%- endif -%}
    {%- set loop_messages = messages[1:] -%}
{%- else -%}
    {%- set first_user_prefix =  -%}
    {%- set loop_messages = messages -%}
{%- endif -%}
{%- for message in loop_messages -%}
    {%- if (message['role'] == 'user') != (loop.index0 % 2 == 0) -%}
        {{ raise_exception("""

# gpt-oss template
CHAT_TEMPLATES[
    "gpt-oss"
] = """{#-
  In addition to the normal inputs of `messages` and `tools`, this template also accepts the
  following kwargs:
  - """

# gptoss template
CHAT_TEMPLATES[
    "gptoss"
] = """{#-
  In addition to the normal inputs of `messages` and `tools`, this template also accepts the
  following kwargs:
  - """

# llama template
CHAT_TEMPLATES["llama"] = (
    """{% if messages[0]['role'] == 'system' %}{% if messages[1]['role'] == 'user' %}{{ bos_token + '[INST] <<SYS>>\n' + messages[0]['content'] + '\n<</SYS>>\n\n' + messages[1]['content'] + ' [/INST]' }}{% set loop_messages = messages[2:] %}{% else %}{{ bos_token + '[INST] ' + messages[0]['content'] + ' [/INST]' }}{% set loop_messages = messages[1:] %}{% endif %}{% else %}{% set loop_messages = messages %}{% endif %}{% for message in loop_messages %}{% if message['role'] == 'user' %}{{ bos_token + '[INST] ' + message['content'].strip() + ' [/INST]' }}{% elif message['role'] == 'assistant' %}{{ ' ' + message['content'].strip() + ' ' + eos_token }}{% else %}{{ raise_exception('Only user and assistant roles are supported!') }}{% endif %}{% endfor %}"""
)

# llama-3 template
CHAT_TEMPLATES["llama-3"] = (
    """{{ bos_token }}{% for message in messages %}{% if message['role'] == 'user' %}{{ '<|start_header_id|>user<|end_header_id|>\n\n' + message['content'] | trim + '<|eot_id|>' }}{% elif message['role'] == 'assistant' %}{{ '<|start_header_id|>assistant<|end_header_id|>\n\n' + message['content'] | trim + '<|eot_id|>' }}{% else %}{{ '<|start_header_id|>' + message['role'] + '<|end_header_id|>\n\n' + message['content'] | trim + '<|eot_id|>' }}{% endif %}{% endfor %}{% if add_generation_prompt %}{{ '<|start_header_id|>assistant<|end_header_id|>\n\n' }}{% endif %}"""
)

# llama-3.1 template
CHAT_TEMPLATES[
    "llama-3.1"
] = """{{- bos_token }}
{%- if custom_tools is defined %}
    {%- set tools = custom_tools %}
{%- endif %}
{%- if not tools_in_user_message is defined %}
    {%- set tools_in_user_message = true %}
{%- endif %}
{%- if not date_string is defined %}
    {%- set date_string = """

# llama-31 template
CHAT_TEMPLATES[
    "llama-31"
] = """{{- bos_token }}
{%- if custom_tools is defined %}
    {%- set tools = custom_tools %}
{%- endif %}
{%- if not tools_in_user_message is defined %}
    {%- set tools_in_user_message = true %}
{%- endif %}
{%- if not date_string is defined %}
    {%- set date_string = """

# llama3 template
CHAT_TEMPLATES["llama3"] = (
    """{{ bos_token }}{% for message in messages %}{% if message['role'] == 'user' %}{{ '<|start_header_id|>user<|end_header_id|>\n\n' + message['content'] | trim + '<|eot_id|>' }}{% elif message['role'] == 'assistant' %}{{ '<|start_header_id|>assistant<|end_header_id|>\n\n' + message['content'] | trim + '<|eot_id|>' }}{% else %}{{ '<|start_header_id|>' + message['role'] + '<|end_header_id|>\n\n' + message['content'] | trim + '<|eot_id|>' }}{% endif %}{% endfor %}{% if add_generation_prompt %}{{ '<|start_header_id|>assistant<|end_header_id|>\n\n' }}{% endif %}"""
)

# mistral template
CHAT_TEMPLATES["mistral"] = (
    """{{ bos_token }}{% if messages[0]['role'] == 'system' %}{% if messages[1]['role'] == 'user' %}{{ '[INST] ' + messages[0]['content'] + ' ' + messages[1]['content'] + ' [/INST]' }}{% set loop_messages = messages[2:] %}{% else %}{{ '[INST] ' + messages[0]['content'] + ' [/INST]' }}{% set loop_messages = messages[1:] %}{% endif %}{% else %}{% set loop_messages = messages %}{% endif %}{% for message in loop_messages %}{% if message['role'] == 'user' %}{{ '[INST] ' + message['content'] + ' [/INST]' }}{% elif message['role'] == 'assistant' %}{{ message['content'] + eos_token }}{% else %}{{ raise_exception('Only user and assistant roles are supported!') }}{% endif %}{% endfor %}"""
)

# phi-3 template
CHAT_TEMPLATES["phi-3"] = (
    """{% for message in messages %}{% if message['role'] == 'user' %}{{'<|user|>\n' + message['content'] + '<|end|>\n'}}{% elif message['role'] == 'assistant' %}{{'<|assistant|>\n' + message['content'] + '<|end|>\n'}}{% else %}{{'<|' + message['role'] + '|>\n' + message['content'] + '<|end|>\n'}}{% endif %}{% endfor %}{% if add_generation_prompt %}{{ '<|assistant|>\n' }}{% endif %}"""
)

# phi-4 template
CHAT_TEMPLATES["phi-4"] = (
    """{% for message in messages %}{% if (message['role'] == 'system') %}{{'<|im_start|>system<|im_sep|>' + message['content'] + '<|im_end|>'}}{% elif (message['role'] == 'user') %}{{'<|im_start|>user<|im_sep|>' + message['content'] + '<|im_end|>'}}{% elif (message['role'] == 'assistant') %}{{'<|im_start|>assistant<|im_sep|>' + message['content'] + '<|im_end|>'}}{% endif %}{% endfor %}{% if add_generation_prompt %}{{ '<|im_start|>assistant<|im_sep|>' }}{% endif %}"""
)

# qwen-2.5 template
CHAT_TEMPLATES[
    "qwen-2.5"
] = """{%- if tools %}
    {{- \'<|im_start|>system\\n\' }}
    {%- if messages[0][\'role\'] == \'system\' %}
        {{- messages[0][\'content\'] }}
    {%- else %}
        {{- \'You are Qwen, created by Alibaba Cloud. You are a helpful assistant.\' }}
    {%- endif %}
    {{- """

# qwen-25 template
CHAT_TEMPLATES[
    "qwen-25"
] = """{%- if tools %}
    {{- \'<|im_start|>system\\n\' }}
    {%- if messages[0][\'role\'] == \'system\' %}
        {{- messages[0][\'content\'] }}
    {%- else %}
        {{- \'You are Qwen, created by Alibaba Cloud. You are a helpful assistant.\' }}
    {%- endif %}
    {{- """

# qwen-3 template
CHAT_TEMPLATES[
    "qwen-3"
] = """
{%- if tools %}
    {{- '<|im_start|>system\n' }}
    {%- if messages[0].role == 'system' %}
        {{- messages[0].content + '\n\n' }}
    {%- endif %}
    {{- """

# qwen2.5 template
CHAT_TEMPLATES[
    "qwen2.5"
] = """{%- if tools %}
    {{- \'<|im_start|>system\\n\' }}
    {%- if messages[0][\'role\'] == \'system\' %}
        {{- messages[0][\'content\'] }}
    {%- else %}
        {{- \'You are Qwen, created by Alibaba Cloud. You are a helpful assistant.\' }}
    {%- endif %}
    {{- """

# qwen25 template
CHAT_TEMPLATES[
    "qwen25"
] = """{%- if tools %}
    {{- \'<|im_start|>system\\n\' }}
    {%- if messages[0][\'role\'] == \'system\' %}
        {{- messages[0][\'content\'] }}
    {%- else %}
        {{- \'You are Qwen, created by Alibaba Cloud. You are a helpful assistant.\' }}
    {%- endif %}
    {{- """

# qwen3 template
CHAT_TEMPLATES[
    "qwen3"
] = """
{%- if tools %}
    {{- '<|im_start|>system\n' }}
    {%- if messages[0].role == 'system' %}
        {{- messages[0].content + '\n\n' }}
    {%- endif %}
    {{- """

# qwen3-instruct template
CHAT_TEMPLATES[
    "qwen3-instruct"
] = """# Tools\\n\\nYou may call one or more functions to assist with the user query.\\n\\nYou are provided with function signatures within <tools></tools> XML tags:\\n<tools>\\n\\n</tools>\\n\\nFor each function call, return a json object with function name and arguments within <tool_call></tool_call> XML tags:\\n<tool_call>\\n{\\: <function-name>, \\: <args-json-object>}\\n</tool_call><|im_end|>\\nuserusersystemassistantname' }}
                {{- tool_call.name }}
                {{- 'argumentstooltooltool"""

# qwen3-thinking template
CHAT_TEMPLATES[
    "qwen3-thinking"
] = """# Tools\\n\\nYou may call one or more functions to assist with the user query.\\n\\nYou are provided with function signatures within <tools></tools> XML tags:\\n<tools>\\n\\n</tools>\\n\\nFor each function call, return a json object with function name and arguments within <tool_call></tool_call> XML tags:\\n<tool_call>\\n{\\: <function-name>, \\: <args-json-object>}\\n</tool_call><|im_end|>\\nuserusersystemassistantname' }}
                {{- tool_call.name }}
                {{- 'argumentstooltooltool"""

# starling template
CHAT_TEMPLATES[
    "starling"
] = """{{ bos_token }}
{%- for message in messages %}
    {{ 'GPT4 Correct ' + message['role'].title() + ': ' + message['content'] + '<|end_of_turn|>' }}
{%- endfor %}
{%- if add_generation_prompt %}
    {{ 'GPT4 Correct Assistant:' }}
{%- endif %}"""

# unsloth template
CHAT_TEMPLATES["unsloth"] = (
    """{{ bos_token }}{% if messages[0]['role'] == 'system' %}{{ messages[0]['content'] + '\n' }}{% set loop_messages = messages[1:] %}{% else %}{{ '{system_message}' + '\n' }}{% set loop_messages = messages %}{% endif %}{% for message in loop_messages %}{% if message['role'] == 'user' %}{{ '>>> User: ' + message['content'] + '\n' }}{% elif message['role'] == 'assistant' %}{{ '>>> Assistant: ' + message['content'] + eos_token + '\n' }}{% else %}{{ raise_exception('Only user and assistant roles are supported!') }}{% endif %}{% endfor %}{% if add_generation_prompt %}{{ '>>> Assistant: ' }}{% endif %}"""
)

# vicuna template
CHAT_TEMPLATES["vicuna"] = (
    """{{ bos_token }}{% if messages[0]['role'] == 'system' %}{{ messages[0]['content'] + ' ' }}{% set loop_messages = messages[1:] %}{% else %}{{ '{system_message}' + ' ' }}{% set loop_messages = messages %}{% endif %}{% for message in loop_messages %}{% if message['role'] == 'user' %}{{ 'USER: ' + message['content'] + ' ' }}{% elif message['role'] == 'assistant' %}{{ 'ASSISTANT: ' + message['content'] + eos_token }}{% else %}{{ raise_exception('Only user and assistant roles are supported!') }}{% endif %}{% endfor %}{% if add_generation_prompt %}{{ 'ASSISTANT:' }}{% endif %}"""
)

# vicuna_old template
CHAT_TEMPLATES["vicuna_old"] = (
    """{{ bos_token }}{% if messages[0]['role'] == 'system' %}{{ messages[0]['content'] + '\n' }}{% set loop_messages = messages[1:] %}{% else %}{{ '{system_message}' + '\n' }}{% set loop_messages = messages %}{% endif %}{% for message in loop_messages %}{% if message['role'] == 'user' %}{{ '### Human: ' + message['content'] + '\n' }}{% elif message['role'] == 'assistant' %}{{ '### Assistant: ' + message['content'] + eos_token + '\n' }}{% else %}{{ raise_exception('Only user and assistant roles are supported!') }}{% endif %}{% endfor %}{% if add_generation_prompt %}{{ '### Assistant:' }}{% endif %}"""
)

# yi-chat template
CHAT_TEMPLATES[
    "yi-chat"
] = """
{% if not add_generation_prompt is defined %}{% set add_generation_prompt = false %}{% endif %}{% for message in messages %}{{'<|im_start|>' + message['role'] + '
' + message['content'] + '<|im_end|>' + '
'}}{% endfor %}{% if add_generation_prompt %}{{ '<|im_start|>assistant
' }}{% endif %}
"""

# zephyr template
CHAT_TEMPLATES["zephyr"] = (
    """{% for message in messages %}{% if message['role'] == 'user' %}{{ '<|user|>\n' + message['content'] + eos_token + '\n' }}{% elif message['role'] == 'assistant' %}{{ '<|assistant|>\n' + message['content'] + eos_token + '\n' }}{% else %}{{ '<|system|>\n' + message['content'] + eos_token + '\n' }}{% endif %}{% endfor %}{% if add_generation_prompt %}{{ '<|assistant|>\n' }}{% endif %}"""
)


def get_chat_template(template_name):
    """
    Get a chat template by name.

    Args:
        template_name: Name of the template (e.g., 'llama3', 'chatml', 'alpaca')

    Returns:
        The template string if found, None otherwise
    """
    return CHAT_TEMPLATES.get(template_name)


def list_templates():
    """
    List all available template names.

    Returns:
        Sorted list of template names
    """
    return sorted(CHAT_TEMPLATES.keys())


def get_template_for_model(model_name):
    """
    Get suggested template based on model name.

    Args:
        model_name: Name or path of the model

    Returns:
        Suggested template name or None
    """
    model_lower = model_name.lower() if model_name else ""

    # Direct matches
    if "llama-3" in model_lower or "llama3" in model_lower:
        return "llama3"
    elif "llama-2" in model_lower or "llama2" in model_lower:
        return "llama"
    elif "gemma-2" in model_lower or "gemma2" in model_lower:
        return "gemma2" if "gemma2" in CHAT_TEMPLATES else "gemma"
    elif "gemma" in model_lower:
        return "gemma"
    elif "mistral" in model_lower:
        return "mistral"
    elif "phi-3" in model_lower or "phi3" in model_lower:
        return "phi-3"
    elif "phi-4" in model_lower or "phi4" in model_lower:
        return "phi-4"
    elif "qwen2.5" in model_lower or "qwen-2.5" in model_lower:
        return "qwen2.5"
    elif "qwen" in model_lower:
        return "qwen3"
    elif "vicuna" in model_lower:
        return "vicuna"
    elif "alpaca" in model_lower:
        return "alpaca"
    elif "chatml" in model_lower:
        return "chatml"
    elif "zephyr" in model_lower:
        return "zephyr"
    elif "yi-" in model_lower and "chat" in model_lower:
        return "yi-chat"

    return None



"""Test that HF tokens are properly passed to model loading functions.

This ensures users can load private HuggingFace models via the web UI.
"""

from unittest.mock import MagicMock, Mock, patch

import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient

from autotrain.app.api_routes import api_router


# Create test app
app = FastAPI()
app.include_router(api_router, prefix="/api")
client = TestClient(app)


@pytest.fixture
def mock_auth(monkeypatch):
    """Mock authentication to allow local access."""
    monkeypatch.delenv("HF_TOKEN", raising=False)


class TestHFTokenPassing:
    """Test that user-provided HF tokens are passed to model loading."""

    def test_token_passed_to_pipeline_for_seq2seq(self, mock_auth, monkeypatch):
        """Verify token from Authorization header is passed to pipeline()."""
        monkeypatch.setattr("autotrain.app.api_routes.detect_model_type", lambda x: "seq2seq")
        monkeypatch.setattr("autotrain.app.api_routes.validate_model_path", lambda x: x)

        captured_token = {}

        # Mock pipeline to capture the token parameter
        def mock_pipeline(task, model, device, token=None):
            captured_token["value"] = token
            mock_pipe = Mock()
            mock_pipe.return_value = [{"generated_text": "test output"}]
            return mock_pipe

        monkeypatch.setattr("autotrain.app.api_routes.pipeline", mock_pipeline)
        monkeypatch.setattr("autotrain.app.api_routes.torch", Mock(cuda=Mock(is_available=Mock(return_value=False))))

        # Make request with Authorization header containing HF token
        test_token = "hf_test_private_model_token_12345"
        request_data = {
            "model_id": "username/private-seq2seq-model",
            "inputs": {"text": "test"},
            "parameters": {"max_new_tokens": 50},
        }

        response = client.post(
            "/api/inference/universal", json=request_data, headers={"Authorization": f"Bearer {test_token}"}
        )

        print(f"\n{'='*80}")
        print("TEST: HF Token Passing to pipeline()")
        print("=" * 80)
        print(f"Sent token in header: {test_token}")
        print(f"Token received by pipeline(): {captured_token.get('value')}")
        print(f"Response status: {response.status_code}")

        assert response.status_code == 200
        assert captured_token["value"] == test_token, f"Expected {test_token}, got {captured_token.get('value')}"

        print("✅ Token successfully passed to pipeline()")
        print("=" * 80)

    def test_token_passed_to_sentence_transformer(self, mock_auth, monkeypatch):
        """Verify token is passed to SentenceTransformer()."""
        monkeypatch.setattr("autotrain.app.api_routes.detect_model_type", lambda x: "sentence-transformers")
        monkeypatch.setattr("autotrain.app.api_routes.validate_model_path", lambda x: x)

        captured_token = {}

        # Mock SentenceTransformer
        class MockSentenceTransformer:
            def __init__(self, model_path, token=None):
                captured_token["value"] = token

            def to(self, device):
                return self

            def encode(self, texts):
                import numpy as np

                if isinstance(texts, str):
                    texts = [texts]
                return np.random.rand(len(texts), 384)

        # Patch at the actual import location (sentence_transformers module)
        monkeypatch.setattr("sentence_transformers.SentenceTransformer", MockSentenceTransformer)

        test_token = "hf_test_private_embeddings_token_67890"
        request_data = {"model_id": "username/private-embeddings-model", "inputs": {"texts": "test sentence"}}

        response = client.post(
            "/api/inference/universal", json=request_data, headers={"Authorization": f"Bearer {test_token}"}
        )

        print(f"\n{'='*80}")
        print("TEST: HF Token Passing to SentenceTransformer()")
        print("=" * 80)
        print(f"Sent token in header: {test_token}")
        print(f"Token received by SentenceTransformer(): {captured_token.get('value')}")
        print(f"Response status: {response.status_code}")

        assert response.status_code == 200
        assert captured_token["value"] == test_token

        print("✅ Token successfully passed to SentenceTransformer()")
        print("=" * 80)

    def test_token_passed_to_question_answering(self, mock_auth, monkeypatch):
        """Verify token is passed to question-answering pipeline."""
        monkeypatch.setattr("autotrain.app.api_routes.detect_model_type", lambda x: "extractive-question-answering")
        monkeypatch.setattr("autotrain.app.api_routes.validate_model_path", lambda x: x)

        captured_token = {}

        def mock_pipeline(task, model, device, token=None):
            captured_token["value"] = token
            mock_pipe = Mock()
            mock_pipe.return_value = {"answer": "test answer", "score": 0.9}
            return mock_pipe

        monkeypatch.setattr("autotrain.app.api_routes.pipeline", mock_pipeline)
        monkeypatch.setattr("autotrain.app.api_routes.torch", Mock(cuda=Mock(is_available=Mock(return_value=False))))

        test_token = "hf_test_private_qa_model_token_99999"
        request_data = {
            "model_id": "username/private-qa-model",
            "inputs": {"text": "Context here", "question": "What is the answer?"},
        }

        response = client.post(
            "/api/inference/universal", json=request_data, headers={"Authorization": f"Bearer {test_token}"}
        )

        print(f"\n{'='*80}")
        print("TEST: HF Token Passing to QA pipeline()")
        print("=" * 80)
        print(f"Sent token in header: {test_token}")
        print(f"Token received by pipeline(): {captured_token.get('value')}")

        assert response.status_code == 200
        assert captured_token["value"] == test_token

        print("✅ Token successfully passed to QA pipeline()")
        print("=" * 80)

    def test_no_token_still_works(self, mock_auth, monkeypatch):
        """Verify inference still works without token (public models)."""
        monkeypatch.setattr("autotrain.app.api_routes.detect_model_type", lambda x: "seq2seq")
        monkeypatch.setattr("autotrain.app.api_routes.validate_model_path", lambda x: x)

        captured_token = {}

        def mock_pipeline(task, model, device, token=None):
            captured_token["value"] = token
            mock_pipe = Mock()
            mock_pipe.return_value = [{"generated_text": "public model output"}]
            return mock_pipe

        monkeypatch.setattr("autotrain.app.api_routes.pipeline", mock_pipeline)
        monkeypatch.setattr("autotrain.app.api_routes.torch", Mock(cuda=Mock(is_available=Mock(return_value=False))))

        request_data = {"model_id": "t5-small", "inputs": {"text": "test"}, "parameters": {"max_new_tokens": 50}}

        # No Authorization header - public model
        response = client.post("/api/inference/universal", json=request_data)

        print(f"\n{'='*80}")
        print("TEST: Public Model (No Token)")
        print("=" * 80)
        print(f"No Authorization header sent")
        print(f"Token received by pipeline(): {captured_token.get('value')}")

        assert response.status_code == 200
        assert captured_token["value"] is None  # Should be None for public models

        print("✅ Public models work without token")
        print("=" * 80)

    def test_token_extraction_from_header(self, mock_auth, monkeypatch):
        """Test that token is correctly extracted from Authorization header."""
        monkeypatch.setattr("autotrain.app.api_routes.detect_model_type", lambda x: "text-classification")
        monkeypatch.setattr("autotrain.app.api_routes.validate_model_path", lambda x: x)

        captured_token = {}

        def mock_pipeline(task, model, device, token=None):
            captured_token["value"] = token
            mock_pipe = Mock()
            mock_pipe.return_value = [{"label": "POSITIVE", "score": 0.99}]
            return mock_pipe

        monkeypatch.setattr("autotrain.app.api_routes.pipeline", mock_pipeline)
        monkeypatch.setattr("autotrain.app.api_routes.torch", Mock(cuda=Mock(is_available=Mock(return_value=False))))

        # Test various token formats
        test_cases = [
            ("Bearer hf_abc123", "hf_abc123"),
            ("Bearer hf_XYZ789_long_token_here", "hf_XYZ789_long_token_here"),
        ]

        for auth_header, expected_token in test_cases:
            # Clear the model cache to ensure pipeline is called for each test case
            import autotrain.app.api_routes

            autotrain.app.api_routes.MODEL_CACHE.clear()

            # Reset the captured token for each test case
            captured_token["value"] = None

            request_data = {"model_id": "test-model", "inputs": {"text": "test"}}

            response = client.post(
                "/api/inference/universal", json=request_data, headers={"Authorization": auth_header}
            )

            print(f"\nAuth header: {auth_header}")
            print(f"Extracted token: {captured_token.get('value')}")

            assert response.status_code == 200
            assert (
                captured_token["value"] == expected_token
            ), f"Expected {expected_token}, got {captured_token.get('value')}"

        print("✅ Token extraction works correctly")


if __name__ == "__main__":
    pytest.main([__file__, "-v", "-s"])



"""Shared catalog metadata for popular models and datasets."""

from __future__ import annotations

import functools
import warnings
from dataclasses import dataclass
from typing import Dict, List, Optional

from autotrain import logger
from autotrain.cli.trainer_metadata import TRAINER_METADATA


#
# Note: We deliberately avoid importing any FastAPI modules at import-time so
# this catalog remains usable from CLI/TUI contexts. We prefer huggingface_hub
# for live listings and fall back to small curated lists when offline.
#


@dataclass(frozen=True)
class CatalogEntry:
    """Simple container for model/dataset metadata."""

    id: str
    label: str
    description: str = ""
    params: Optional[int] = None


def format_params(num_params: Optional[int]) -> str:
    """Format parameter count into human-readable string."""
    if not num_params:
        return ""
    if num_params >= 1e12:
        return f"({num_params/1e12:.1f}T)"
    elif num_params >= 1e9:
        return f"({num_params/1e9:.1f}B)"
    elif num_params >= 1e6:
        return f"({num_params/1e6:.0f}M)"
    else:
        return f"({num_params/1e3:.0f}K)"


def _trainer_key(trainer_type: Optional[str], trainer_variant: Optional[str]) -> str:
    """Normalize trainer identifiers to a catalog key."""
    if not trainer_type:
        return "llm:sft"
    if trainer_type == "llm":
        variant = trainer_variant or "sft"
        return f"llm:{variant}"
    return trainer_type


def _model_catalog_fallback(trainer_key: str) -> List[CatalogEntry]:
    """Return static fallback models when dynamic fetch is unavailable."""
    if trainer_key.startswith("llm:dpo"):
        entries = [
            CatalogEntry("mistralai/Mistral-7B-Instruct-v0.2", "Mistral 7B Instruct"),
            CatalogEntry("meta-llama/Llama-3.1-8B-Instruct", "Llama 3.1 8B Instruct"),
            CatalogEntry("HuggingFaceH4/zephyr-7b-beta", "Zephyr 7B Beta"),
        ]
    elif trainer_key.startswith("llm:ppo"):
        entries = [
            CatalogEntry("meta-llama/Llama-3.2-1B", "Llama 3.2 1B"),
            CatalogEntry("Qwen/Qwen2.5-0.5B", "Qwen2.5 0.5B"),
        ]
    elif trainer_key.startswith("llm"):
        entries = [
            CatalogEntry("google/gemma-3-270m", "Gemma 3 270M (default)"),
            CatalogEntry("meta-llama/Llama-3.2-1B", "Llama 3.2 1B"),
            CatalogEntry("meta-llama/Llama-3.2-3B", "Llama 3.2 3B"),
            CatalogEntry("Qwen/Qwen2.5-0.5B", "Qwen2.5 0.5B"),
            CatalogEntry("HuggingFaceTB/SmolLM2-360M", "SmolLM2 360M"),
        ]
    else:
        metadata = TRAINER_METADATA.get(trainer_key)
        starter = metadata.get("starter_models", []) if metadata else []
        entries = [CatalogEntry(model_id, model_id) for model_id in starter]
    return entries


def _dataset_catalog_fallback(trainer_key: str) -> List[CatalogEntry]:
    """Static curated datasets for each trainer."""
    catalog: Dict[str, List[CatalogEntry]] = {
        "llm:sft": [
            CatalogEntry("tatsu-lab/alpaca", "Stanford Alpaca", "52k instruction-following samples"),
            CatalogEntry("yahma/alpaca-cleaned", "Alpaca Cleaned", "Cleaner variant of Alpaca"),
            CatalogEntry("HuggingFaceH4/ultrachat_200k", "UltraChat 200k", "Long-form chat prompts"),
            CatalogEntry("Open-Orca/SlimOrca", "SlimOrca", "Preference distilled Orca data"),
            CatalogEntry("LDJnr/Puffin", "Puffin", "English instruction fine-tuning"),
        ],
        "llm:dpo": [
            CatalogEntry(
                "argilla/ultrafeedback-binarized-preferences", "UltraFeedback", "Preference pairs for DPO/ORPO"
            ),
            CatalogEntry("Anthropic/hh-rlhf", "HH RLHF", "Helpful/Harmless preference data"),
            CatalogEntry("OpenAssistant/oasst_top1_2023-08-25", "OpenAssistant Top-1", "Human preference picks"),
        ],
        "llm:orpo": [
            CatalogEntry("argilla/ultrafeedback-binarized-preferences", "UltraFeedback", "Preference pairs for ORPO"),
            CatalogEntry("mosaicml/instruct-v3", "Mosaic Instruct", "Paired responses for preference training"),
        ],
        "llm:ppo": [
            CatalogEntry("Open-Orca/SlimOrca", "SlimOrca", "Pairs suitable for reward modeling"),
            CatalogEntry("Intel/orca_dpo_pairs", "Orca DPO Pairs", "Short preference prompts"),
            CatalogEntry("openbmb/UltraInteract_pair", "UltraInteract", "Conversational preference data"),
        ],
        "text-classification": [
            CatalogEntry("ag_news", "AG News", "News topic classification"),
            CatalogEntry("yelp_review_full", "Yelp Review Full", "5-class sentiment"),
            CatalogEntry("dbpedia_14", "DBPedia 14", "Ontology classification"),
        ],
        "token-classification": [
            CatalogEntry("conll2003", "CoNLL 2003", "NER dataset"),
            CatalogEntry("wnut_17", "WNUT 2017", "Emerging entities"),
        ],
        "tabular": [
            CatalogEntry("autogluon/ames", "Ames Housing", "Regression benchmark"),
            CatalogEntry("autogluon/otto", "Otto Group", "Multi-class classification"),
        ],
        "image-classification": [
            CatalogEntry("huggingface/cifar10", "CIFAR-10", "10-class images"),
            CatalogEntry("huggingface/beans", "Beans", "Plant disease detection"),
        ],
        "image-regression": [
            CatalogEntry("Shenggan/BraTS-GLIOMA", "BraTS Glioma", "Medical regression labels"),
        ],
        "extractive-qa": [
            CatalogEntry("squad", "SQuAD v1.1", "Reading comprehension"),
            CatalogEntry("rajpurkar/squad_v2", "SQuAD v2", "Q&A with unanswerables"),
        ],
        "seq2seq": [
            CatalogEntry("cnn_dailymail", "CNN/DailyMail", "Summarization"),
            CatalogEntry("wmt16", "WMT16", "Machine translation corpus"),
        ],
        "sent-transformers": [
            CatalogEntry("sentence-transformers/all-nli", "AllNLI", "STS + NLI pairs"),
        ],
        "vlm": [
            CatalogEntry("laion/OIG", "OIG", "Open Instruction Generalist prompts"),
            CatalogEntry("llava-bench-instruct/llava-bench-instruct", "LLaVA-Bench", "VLM instructions"),
        ],
    }
    return catalog.get(trainer_key, [])


def _map_trainer_to_model_pipeline(trainer_type: Optional[str], trainer_variant: Optional[str]) -> Optional[str]:
    """Best-effort mapping from trainer to HF model pipeline_tag."""
    if not trainer_type:
        return "text-generation"
    if trainer_type == "llm":
        # SFT/DPO/ORPO/Reward/PPO all ultimately fine-tune text-generation models
        return "text-generation"
    return {
        "text-classification": "text-classification",
        "token-classification": "token-classification",
        "seq2seq": "summarization",  # heuristic; could be translation too
        "extractive-qa": "question-answering",
        "sent-transformers": "sentence-similarity",
        "vlm": "image-text-to-text",
        "image-classification": "image-classification",
        "image-regression": "image-classification",  # no direct regression pipeline in hub listing
        "tabular": None,
    }.get(trainer_type, None)


def _map_trainer_to_dataset_task(trainer_type: Optional[str], trainer_variant: Optional[str]) -> Optional[str]:
    """Best-effort mapping from trainer to HF datasets task_categories filter."""
    if not trainer_type:
        return None
    if trainer_type == "llm":
        # No clear 'text-generation' dataset category; keep curated fallback for LLM
        return None
    return {
        "text-classification": "text-classification",
        "token-classification": "token-classification",
        "seq2seq": "summarization",  # heuristic; "translation" also valid depending on use
        "extractive-qa": "question-answering",
        "sent-transformers": "sentence-similarity",
        "vlm": "visual-question-answering",
        "image-classification": "image-classification",
        "image-regression": "image-classification",
        "tabular": None,
    }.get(trainer_type, None)


def _fetch_trending(repo_type: str, limit: int = 20) -> List[CatalogEntry]:
    """Fetch trending models/datasets from Hugging Face."""
    try:
        import requests

        resp = requests.get(f"https://huggingface.co/api/trending?type={repo_type}&limit={limit}", timeout=10)
        resp.raise_for_status()
        data = resp.json()
        entries: List[CatalogEntry] = []
        # The API returns a list of objects, sometimes nested
        current_list = data.get("recentlyTrending") if isinstance(data, dict) else data

        for item in current_list:
            repo_id = item.get("repoData", {}).get("repoId") or item.get("repoId")
            if not repo_id:
                continue

            # Extract useful info
            params = item.get("repoData", {}).get("numParameters") or item.get("numParameters")

            entries.append(CatalogEntry(repo_id, repo_id, "", params))
            if len(entries) >= limit:
                break
        return entries
    except Exception as exc:
        logger.debug(f"Unable to fetch trending {repo_type}: {exc}")
        return []


def _hf_list_top_models(
    trainer_type: Optional[str],
    trainer_variant: Optional[str],
    limit: int = 20,
    sort_by: str = "downloads",
    search_query: Optional[str] = None,
) -> List[CatalogEntry]:
    """Fetch top models from Hugging Face hub via huggingface_hub."""
    try:
        from huggingface_hub import HfApi

        pipeline = _map_trainer_to_model_pipeline(trainer_type, trainer_variant)
        # If searching, we might want to relax pipeline filter if it restricts too much,
        # but keeping it ensures relevant models.

        api = HfApi()

        # Map sort_by to HfApi sort
        sort_key = sort_by
        direction = -1

        if sort_by == "trending":
            # HfApi doesn't support trending sort directly, use the trending API
            return _fetch_trending("model", limit)

        if sort_by == "recent":
            sort_key = "createdAt"
        elif sort_by == "modified":
            sort_key = "lastModified"

        args = {
            "sort": sort_key,
            "direction": direction,
            "limit": limit,
        }

        if pipeline:
            args["filter"] = pipeline

        if search_query:
            args["search"] = search_query

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=FutureWarning, module="huggingface_hub.utils._deprecation")
            results = api.list_models(**args)
        entries: List[CatalogEntry] = []
        # results may be ModelInfo objects
        for r in results:
            model_id = getattr(r, "modelId", None) or getattr(r, "id", None) or str(r)
            if not model_id:
                continue

            # Try to get params if available (ModelInfo might have it in safetensors metadata)
            # but HfApi list_models might not return full details.
            # We'll accept it might be None.
            params = None
            # getattr(r, "safetensors", {}) ...

            entries.append(CatalogEntry(model_id, model_id, "", params))
        return entries
    except Exception as exc:
        logger.debug(f"Unable to fetch models from HF hub: {exc}")
        return []


def _hf_list_top_datasets(
    trainer_type: Optional[str],
    trainer_variant: Optional[str],
    limit: int = 20,
    sort_by: str = "downloads",
    search_query: Optional[str] = None,
) -> List[CatalogEntry]:
    """Fetch top datasets from Hugging Face hub via raw REST API (task filter)."""
    try:
        # Use requests to call the public REST endpoint for datasets
        import os

        import requests

        if sort_by == "trending" and not search_query:
            return _fetch_trending("dataset", limit)

        task = _map_trainer_to_dataset_task(trainer_type, trainer_variant)

        params = {
            "sort": sort_by if sort_by != "recent" else "createdAt",
            "limit": str(limit),
        }

        if sort_by == "trending":
            # Fallback if search is used with trending (API doesn't support search+trending easily)
            # We'll just search and sort by downloads
            params["sort"] = "downloads"

        if task:
            params["task_categories"] = task

        if search_query:
            params["search"] = search_query

        headers = {}
        token = os.environ.get("HF_TOKEN")
        if token:
            headers["Authorization"] = f"Bearer {token}"

        resp = requests.get("https://huggingface.co/api/datasets", params=params, headers=headers, timeout=10)
        resp.raise_for_status()
        data = resp.json()
        entries: List[CatalogEntry] = []
        for item in data:
            ds_id = item.get("id")
            if not ds_id:
                continue

            # Dataset size/rows?
            # Usually not in list response
            entries.append(CatalogEntry(ds_id, ds_id))
        return entries
    except Exception as exc:
        logger.debug(f"Unable to fetch datasets from HF hub: {exc}")
        return []


@functools.lru_cache(maxsize=64)
def get_popular_models(
    trainer_type: Optional[str],
    trainer_variant: Optional[str] = None,
    sort_by: str = "trending",
    search_query: Optional[str] = None,
) -> List[CatalogEntry]:
    """Return curated model IDs for the given trainer, preferring hub metadata when available."""
    trainer_key = _trainer_key(trainer_type, trainer_variant)

    # 1) Try HF hub directly (works in CLI/TUI without FastAPI)
    entries = _hf_list_top_models(trainer_type, trainer_variant, limit=20, sort_by=sort_by, search_query=search_query)
    if entries:
        return entries

    # 2) Try UI-centric model map if available (FastAPI context) - Only if no search/sort
    if sort_by == "trending" and not search_query:
        try:
            from autotrain.app.models import fetch_models  # optional

            model_map = fetch_models()
            model_key = {
                "text-classification": "text-classification",
                "token-classification": "token-classification",
                "tabular": "tabular-classification",
                "image-classification": "image-classification",
                "image-regression": "image-regression",
                "seq2seq": "seq2seq",
                "extractive-qa": "extractive-qa",
                "sent-transformers": "sentence-transformers",
                "vlm": "vlm",
            }.get(trainer_type or "llm", "llm")
            hub_entries = model_map.get(model_key, [])
            if hub_entries:
                return [CatalogEntry(model_id, model_id) for model_id in hub_entries[:20]]
        except Exception as exc:  # pragma: no cover
            logger.debug(f"Unable to fetch hub model catalog via app.models: {exc}")

    # 3) Fallback curated list
    if not search_query:
        return _model_catalog_fallback(trainer_key)

    return []


@functools.lru_cache(maxsize=64)
def get_popular_datasets(
    trainer_type: Optional[str],
    trainer_variant: Optional[str] = None,
    sort_by: str = "trending",
    search_query: Optional[str] = None,
) -> List[CatalogEntry]:
    """Return curated dataset IDs for the given trainer."""
    trainer_key = _trainer_key(trainer_type, trainer_variant)

    # 1) Try HF hub for supported task categories
    entries = _hf_list_top_datasets(
        trainer_type, trainer_variant, limit=20, sort_by=sort_by, search_query=search_query
    )
    if entries:
        return entries

    # 2) Fallback curated list (only if no search)
    if not search_query:
        entries = _dataset_catalog_fallback(trainer_key)
        if entries:
            return entries
        # 3) Final fallback to TRAINER_METADATA defaults
        if trainer_type and trainer_type in TRAINER_METADATA:
            metadata = TRAINER_METADATA[trainer_type]
            default = metadata.get("default_dataset")
            if default:
                return [CatalogEntry(default, default)]
    return []



class Seq2SeqDataset:
    """
    A dataset class for sequence-to-sequence tasks.

    Args:
        data (list): The dataset containing input and target sequences.
        tokenizer (PreTrainedTokenizer): The tokenizer to process the text data.
        config (object): Configuration object containing dataset parameters.

    Attributes:
        data (list): The dataset containing input and target sequences.
        tokenizer (PreTrainedTokenizer): The tokenizer to process the text data.
        config (object): Configuration object containing dataset parameters.
        max_len_input (int): Maximum length for input sequences.
        max_len_target (int): Maximum length for target sequences.

    Methods:
        __len__(): Returns the number of samples in the dataset.
        __getitem__(item): Returns the tokenized input and target sequences for a given index.
    """

    def __init__(self, data, tokenizer, config):
        self.data = data
        self.tokenizer = tokenizer
        self.config = config
        self.max_len_input = self.config.max_seq_length
        self.max_len_target = self.config.max_target_length

        # Validate required columns exist
        if len(self.data) > 0:
            # Check if data has column_names attribute (datasets library)
            if hasattr(self.data, "column_names"):
                columns = self.data.column_names
            elif hasattr(self.data, "columns"):
                columns = self.data.columns
            else:
                # For list-like data, check the first item
                columns = list(self.data[0].keys()) if isinstance(self.data[0], dict) else []

            # Validate text column exists
            if columns and self.config.text_column not in columns:
                raise KeyError(
                    f"Required column '{self.config.text_column}' not found in dataset. Available columns: {columns}"
                )

            # Validate target column exists
            if columns and self.config.target_column not in columns:
                raise KeyError(
                    f"Required column '{self.config.target_column}' not found in dataset. Available columns: {columns}"
                )

    def __len__(self):
        return len(self.data)

    def __getitem__(self, item):
        text = str(self.data[item][self.config.text_column])
        target = str(self.data[item][self.config.target_column])

        model_inputs = self.tokenizer(text, max_length=self.max_len_input, truncation=True)

        labels = self.tokenizer(text_target=target, max_length=self.max_len_target, truncation=True)

        model_inputs["labels"] = labels["input_ids"]
        return model_inputs



"""
Metadata for all trainer types: FIELD_GROUPS and FIELD_SCOPES
This module centralizes parameter organization and scope information for the interactive wizard.
"""

from autotrain.trainers.extractive_question_answering.params import ExtractiveQuestionAnsweringParams
from autotrain.trainers.image_classification.params import ImageClassificationParams
from autotrain.trainers.image_regression.params import ImageRegressionParams
from autotrain.trainers.sent_transformers.params import SentenceTransformersParams
from autotrain.trainers.seq2seq.params import Seq2SeqParams
from autotrain.trainers.tabular.params import TabularParams
from autotrain.trainers.text_classification.params import TextClassificationParams
from autotrain.trainers.token_classification.params import TokenClassificationParams
from autotrain.trainers.vlm.params import VLMTrainingParams


# ===== TEXT CLASSIFICATION =====

TEXT_CLASSIFICATION_FIELD_GROUPS = {
    "Basic": [
        "model",
        "project_name",
        "data_path",
        "train_split",
        "valid_split",
        "max_samples",
    ],
    "Data Processing": [
        "text_column",
        "target_column",
        "max_seq_length",
    ],
    "Training Configuration": [
        "log",
        "logging_steps",
        "eval_strategy",
        "auto_find_batch_size",
        "mixed_precision",
        "save_total_limit",
        "early_stopping_patience",
        "early_stopping_threshold",
    ],
    "Training Hyperparameters": [
        "lr",
        "epochs",
        "batch_size",
        "warmup_ratio",
        "gradient_accumulation",
        "optimizer",
        "scheduler",
        "weight_decay",
        "max_grad_norm",
        "seed",
    ],
    "Hub Integration": [
        "push_to_hub",
        "username",
        "token",
    ],
}

# ===== TOKEN CLASSIFICATION =====

TOKEN_CLASSIFICATION_FIELD_GROUPS = {
    "Basic": [
        "model",
        "project_name",
        "data_path",
        "train_split",
        "valid_split",
        "max_samples",
    ],
    "Data Processing": [
        "tokens_column",
        "tags_column",
        "max_seq_length",
    ],
    "Training Configuration": [
        "log",
        "logging_steps",
        "eval_strategy",
        "auto_find_batch_size",
        "mixed_precision",
        "save_total_limit",
        "early_stopping_patience",
        "early_stopping_threshold",
    ],
    "Training Hyperparameters": [
        "lr",
        "epochs",
        "batch_size",
        "warmup_ratio",
        "gradient_accumulation",
        "optimizer",
        "scheduler",
        "weight_decay",
        "max_grad_norm",
        "seed",
    ],
    "Hub Integration": [
        "push_to_hub",
        "username",
        "token",
    ],
}

# ===== TABULAR =====

TABULAR_FIELD_GROUPS = {
    "Basic": [
        "model",
        "project_name",
        "data_path",
        "train_split",
        "valid_split",
        "max_samples",
        "seed",
    ],
    "Data Processing": [
        "id_column",
        "target_columns",
        "categorical_columns",
        "numerical_columns",
        "task",
    ],
    "Training Configuration": [
        "num_trials",
        "time_limit",
        "categorical_imputer",
        "numerical_imputer",
        "numeric_scaler",
    ],
    "Hub Integration": [
        "push_to_hub",
        "username",
        "token",
    ],
}

# ===== IMAGE CLASSIFICATION =====

IMAGE_CLASSIFICATION_FIELD_GROUPS = {
    "Basic": [
        "model",
        "project_name",
        "data_path",
        "train_split",
        "valid_split",
        "max_samples",
    ],
    "Data Processing": [
        "image_column",
        "target_column",
    ],
    "Training Configuration": [
        "log",
        "logging_steps",
        "eval_strategy",
        "auto_find_batch_size",
        "mixed_precision",
        "save_total_limit",
        "early_stopping_patience",
        "early_stopping_threshold",
    ],
    "Training Hyperparameters": [
        "lr",
        "epochs",
        "batch_size",
        "warmup_ratio",
        "gradient_accumulation",
        "optimizer",
        "scheduler",
        "weight_decay",
        "max_grad_norm",
        "seed",
    ],
    "Hub Integration": [
        "push_to_hub",
        "username",
        "token",
    ],
}

# ===== IMAGE REGRESSION =====

IMAGE_REGRESSION_FIELD_GROUPS = {
    "Basic": [
        "model",
        "project_name",
        "data_path",
        "train_split",
        "valid_split",
        "max_samples",
    ],
    "Data Processing": [
        "image_column",
        "target_column",
    ],
    "Training Configuration": [
        "log",
        "logging_steps",
        "eval_strategy",
        "auto_find_batch_size",
        "mixed_precision",
        "save_total_limit",
        "early_stopping_patience",
        "early_stopping_threshold",
    ],
    "Training Hyperparameters": [
        "lr",
        "epochs",
        "batch_size",
        "warmup_ratio",
        "gradient_accumulation",
        "optimizer",
        "scheduler",
        "weight_decay",
        "max_grad_norm",
        "seed",
    ],
    "Hub Integration": [
        "push_to_hub",
        "username",
        "token",
    ],
}

# ===== SEQ2SEQ =====

SEQ2SEQ_FIELD_GROUPS = {
    "Basic": [
        "model",
        "project_name",
        "data_path",
        "train_split",
        "valid_split",
        "max_samples",
    ],
    "Data Processing": [
        "text_column",
        "target_column",
        "max_seq_length",
        "max_target_length",
    ],
    "Training Configuration": [
        "log",
        "logging_steps",
        "eval_strategy",
        "auto_find_batch_size",
        "mixed_precision",
        "save_total_limit",
        "early_stopping_patience",
        "early_stopping_threshold",
    ],
    "Training Hyperparameters": [
        "lr",
        "epochs",
        "batch_size",
        "warmup_ratio",
        "gradient_accumulation",
        "optimizer",
        "scheduler",
        "weight_decay",
        "max_grad_norm",
        "seed",
    ],
    "PEFT/LoRA": [
        "peft",
        "quantization",
        "lora_r",
        "lora_alpha",
        "lora_dropout",
        "target_modules",
    ],
    "Hub Integration": [
        "push_to_hub",
        "username",
        "token",
    ],
}

# ===== EXTRACTIVE QA =====

EXTRACTIVE_QA_FIELD_GROUPS = {
    "Basic": [
        "model",
        "project_name",
        "data_path",
        "train_split",
        "valid_split",
        "max_samples",
    ],
    "Data Processing": [
        "text_column",
        "question_column",
        "answer_column",
        "max_seq_length",
        "max_doc_stride",
    ],
    "Training Configuration": [
        "log",
        "logging_steps",
        "eval_strategy",
        "auto_find_batch_size",
        "mixed_precision",
        "save_total_limit",
        "early_stopping_patience",
        "early_stopping_threshold",
    ],
    "Training Hyperparameters": [
        "lr",
        "epochs",
        "batch_size",
        "warmup_ratio",
        "gradient_accumulation",
        "optimizer",
        "scheduler",
        "weight_decay",
        "max_grad_norm",
        "seed",
    ],
    "Hub Integration": [
        "push_to_hub",
        "username",
        "token",
    ],
}

# ===== SENTENCE TRANSFORMERS =====

SENTENCE_TRANSFORMERS_FIELD_GROUPS = {
    "Basic": [
        "model",
        "project_name",
        "data_path",
        "train_split",
        "valid_split",
        "max_samples",
    ],
    "Data Processing": [
        "sentence1_column",
        "sentence2_column",
        "sentence3_column",
        "target_column",
        "max_seq_length",
        "trainer",
    ],
    "Training Configuration": [
        "log",
        "logging_steps",
        "eval_strategy",
        "auto_find_batch_size",
        "mixed_precision",
        "save_total_limit",
        "early_stopping_patience",
        "early_stopping_threshold",
    ],
    "Training Hyperparameters": [
        "lr",
        "epochs",
        "batch_size",
        "warmup_ratio",
        "gradient_accumulation",
        "optimizer",
        "scheduler",
        "weight_decay",
        "max_grad_norm",
        "seed",
    ],
    "Hub Integration": [
        "push_to_hub",
        "username",
        "token",
    ],
}

# ===== VLM (Vision Language Models) =====

VLM_FIELD_GROUPS = {
    "Basic": [
        "model",
        "project_name",
        "data_path",
        "train_split",
        "valid_split",
        "max_samples",
    ],
    "Data Processing": [
        "text_column",
        "image_column",
        "prompt_text_column",
        "trainer",
    ],
    "Training Configuration": [
        "log",
        "logging_steps",
        "eval_strategy",
        "save_total_limit",
        "auto_find_batch_size",
        "mixed_precision",
        "disable_gradient_checkpointing",
    ],
    "Training Hyperparameters": [
        "lr",
        "epochs",
        "batch_size",
        "warmup_ratio",
        "gradient_accumulation",
        "optimizer",
        "scheduler",
        "weight_decay",
        "max_grad_norm",
        "seed",
    ],
    "PEFT/LoRA": [
        "peft",
        "quantization",
        "target_modules",
        "merge_adapter",
        "lora_r",
        "lora_alpha",
        "lora_dropout",
    ],
    "Hub Integration": [
        "push_to_hub",
        "username",
        "token",
    ],
}

# ===== TRAINER METADATA REGISTRY =====

TRAINER_METADATA = {
    "text-classification": {
        "params_class": TextClassificationParams,
        "field_groups": TEXT_CLASSIFICATION_FIELD_GROUPS,
        "display_name": "Text Classification",
        "description": "Classify text into predefined categories",
        "required_columns": ["text_column", "target_column"],
        "default_model": "bert-base-uncased",
        "starter_models": [
            "bert-base-uncased",
            "distilbert-base-uncased",
            "roberta-base",
            "albert-base-v2",
        ],
    },
    "token-classification": {
        "params_class": TokenClassificationParams,
        "field_groups": TOKEN_CLASSIFICATION_FIELD_GROUPS,
        "display_name": "Token Classification (NER)",
        "description": "Label individual tokens (Named Entity Recognition, POS tagging)",
        "required_columns": ["tokens_column", "tags_column"],
        "default_model": "bert-base-uncased",
        "starter_models": [
            "bert-base-uncased",
            "distilbert-base-uncased",
            "roberta-base",
        ],
    },
    "tabular": {
        "params_class": TabularParams,
        "field_groups": TABULAR_FIELD_GROUPS,
        "display_name": "Tabular",
        "description": "Train on structured/tabular data (classification or regression)",
        "required_columns": ["target_columns"],
        "default_model": "xgboost",
        "starter_models": [
            "xgboost",
            "random_forest",
            "lightgbm",
            "catboost",
            "logistic_regression",
            "ridge",
            "svm",
            "extra_trees",
        ],
    },
    "image-classification": {
        "params_class": ImageClassificationParams,
        "field_groups": IMAGE_CLASSIFICATION_FIELD_GROUPS,
        "display_name": "Image Classification",
        "description": "Classify images into predefined categories",
        "required_columns": ["image_column", "target_column"],
        "default_model": "google/vit-base-patch16-224",
        "starter_models": [
            "google/vit-base-patch16-224",
            "microsoft/resnet-50",
            "facebook/convnext-tiny-224",
        ],
    },
    "image-regression": {
        "params_class": ImageRegressionParams,
        "field_groups": IMAGE_REGRESSION_FIELD_GROUPS,
        "display_name": "Image Regression",
        "description": "Predict continuous values from images",
        "required_columns": ["image_column", "target_column"],
        "default_model": "google/vit-base-patch16-224",
        "starter_models": [
            "google/vit-base-patch16-224",
            "microsoft/resnet-50",
        ],
    },
    "seq2seq": {
        "params_class": Seq2SeqParams,
        "field_groups": SEQ2SEQ_FIELD_GROUPS,
        "display_name": "Seq2Seq",
        "description": "Sequence-to-sequence tasks (translation, summarization)",
        "required_columns": ["text_column", "target_column"],
        "default_model": "t5-small",
        "starter_models": [
            "t5-small",
            "t5-base",
            "google/flan-t5-small",
            "facebook/bart-base",
        ],
    },
    "extractive-qa": {
        "params_class": ExtractiveQuestionAnsweringParams,
        "field_groups": EXTRACTIVE_QA_FIELD_GROUPS,
        "display_name": "Extractive Question Answering",
        "description": "Extract answers from context passages",
        "required_columns": ["text_column", "question_column", "answer_column"],
        "default_model": "bert-base-uncased",
        "starter_models": [
            "bert-base-uncased",
            "distilbert-base-uncased",
            "roberta-base",
        ],
    },
    "sent-transformers": {
        "params_class": SentenceTransformersParams,
        "field_groups": SENTENCE_TRANSFORMERS_FIELD_GROUPS,
        "display_name": "Sentence Transformers",
        "description": "Train embeddings for semantic similarity",
        "required_columns": ["sentence1_column"],
        "default_model": "sentence-transformers/all-MiniLM-L6-v2",
        "starter_models": [
            "sentence-transformers/all-MiniLM-L6-v2",
            "sentence-transformers/all-mpnet-base-v2",
            "bert-base-uncased",
        ],
    },
    "vlm": {
        "params_class": VLMTrainingParams,
        "field_groups": VLM_FIELD_GROUPS,
        "display_name": "Vision-Language Model",
        "description": "Train models that understand both images and text",
        "required_columns": ["text_column", "image_column"],
        "default_model": "HuggingFaceM4/idefics2-8b",
        "starter_models": [
            "HuggingFaceM4/idefics2-8b",
            "llava-hf/llava-1.5-7b-hf",
        ],
    },
}


def get_trainer_metadata(trainer_type):
    """Get metadata for a specific trainer type."""
    return TRAINER_METADATA.get(trainer_type)


def get_all_trainer_types():
    """Get list of all available trainer types."""
    return list(TRAINER_METADATA.keys())


def get_trainer_display_name(trainer_type):
    """Get display name for a trainer type."""
    metadata = TRAINER_METADATA.get(trainer_type)
    return metadata["display_name"] if metadata else trainer_type



"""
Shared sweep utilities for CLM trainers
"""

import json
import os
from typing import Any, Callable

from autotrain import logger
from autotrain.trainers.clm.params import LLMTrainingParams


def run_with_sweep(config: LLMTrainingParams, train_func: Callable) -> Any:
    """
    Run hyperparameter sweep for any CLM trainer.

    Args:
        config: Training configuration with sweep parameters
        train_func: The trainer-specific train function to call

    Returns:
        The final trained model with best hyperparameters
    """
    from autotrain.utils import run_autotrain_sweep

    logger.info(f"Running hyperparameter sweep with {config.sweep_backend} backend")

    # Parse sweep parameters from JSON
    if config.sweep_params:
        sweep_params_dict = json.loads(config.sweep_params)
    else:
        # Default sweep parameters
        sweep_params_dict = {
            "lr": (1e-5, 1e-3, "log_uniform"),
            "batch_size": [2, 4, 8, 16],
            "warmup_ratio": (0.0, 0.2, "float"),
        }

    # Create train function for sweep
    def train_for_sweep(params):
        # Set WANDB_PROJECT so trainer's internal wandb.init() uses correct project
        # This prevents the trainer from logging to 'huggingface' default project
        if config.log == "wandb":
            # Use basename of project_name since it's the output path, not a W&B project name
            wandb_project = getattr(config, "wandb_sweep_project", None) or os.path.basename(config.project_name)
            os.environ["WANDB_PROJECT"] = wandb_project
            if getattr(config, "wandb_sweep_entity", None):
                os.environ["WANDB_ENTITY"] = config.wandb_sweep_entity
            logger.info(f"Set WANDB_PROJECT={wandb_project} for trainer")

        # Create new config with sweep params
        trial_config = LLMTrainingParams(**{**vars(config), **params})
        trial_config.use_sweep = False  # Prevent recursive sweep

        # Run training
        trainer = train_func(trial_config)

        # Extract metric from trainer
        if hasattr(trainer, "state") and hasattr(trainer.state, "log_history") and trainer.state.log_history:
            metric_name = config.sweep_metric or "eval_loss"
            for log in reversed(trainer.state.log_history):
                if metric_name in log:
                    return log[metric_name]

        # Return worst possible value if metric not found
        return float("inf") if config.sweep_direction == "minimize" else float("-inf")

    # Run sweep using existing implementation
    result = run_autotrain_sweep(
        model_config=vars(config),
        sweep_parameters=sweep_params_dict,
        train_function=train_for_sweep,
        metric=config.sweep_metric or "eval_loss",
        direction=config.sweep_direction or "minimize",
        n_trials=config.sweep_n_trials,
        backend=config.sweep_backend or "optuna",
        output_dir=f"{config.project_name}/sweep",
        wandb_sweep=getattr(config, "wandb_sweep", False),
        wandb_project=getattr(config, "wandb_sweep_project", None) or os.path.basename(config.project_name),
        wandb_entity=getattr(config, "wandb_sweep_entity", None),
        wandb_sweep_id=getattr(config, "wandb_sweep_id", None),
        post_trial_script=getattr(config, "post_trial_script", None),
    )

    logger.info(f"Sweep completed! Best params: {result.best_params}")
    logger.info(f"Best {config.sweep_metric}: {result.best_value}")

    # Train final model with best parameters
    logger.info("Training final model with best parameters...")
    final_config = LLMTrainingParams(**{**vars(config), **result.best_params})
    final_config.use_sweep = False

    return train_func(final_config)


def with_sweep(train_func: Callable) -> Callable:
    """
    Decorator to add sweep functionality to any trainer.

    Usage:
        @with_sweep
        def train(config):
            # your training logic
            return trainer
    """

    def wrapper(config):
        if hasattr(config, "use_sweep") and config.use_sweep:
            return run_with_sweep(config, train_func)
        return train_func(config)

    return wrapper



"""
Real-world tests for dataset conversion with actual datasets and models.

Tests the full pipeline with real Unsloth integration, no mocks.
"""

import json
import os
import sys

import pandas as pd
import pytest
from datasets import load_dataset
from transformers import AutoTokenizer


# Add parent directory to path
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from autotrain.preprocessor.llm import (
    analyze_and_convert_dataset,
    apply_chat_template,
    convert_alpaca_to_messages,
    convert_dpo_to_messages,
    convert_sharegpt_to_messages,
    detect_dataset_format,
    extend_alpaca_conversations,
    formatting_prompts_func,
    get_available_chat_templates,
    standardize_dataset,
)


# Test configuration
SAMPLE_SIZE = 5
TEST_MODELS = [
    "HuggingFaceH4/zephyr-7b-beta",  # Uses zephyr/chatml template
    "meta-llama/Llama-2-7b-hf",  # Uses llama template
    "google/gemma-2b",  # Uses gemma template
    "microsoft/phi-2",  # Uses phi template
]

TEST_DATASETS = {
    "alpaca": "tatsu-lab/alpaca",
    "sharegpt": "philschmid/guanaco-sharegpt-style",
    "messages": "HuggingFaceH4/no_robots",  # Already in messages format
    "dpo": "argilla/ultrafeedback-binarized-preferences-cleaned",
}


def validate_chat_templates():
    """Validate that chat templates are available and working."""
    try:
        # Try Unsloth first if available
        import torch

        if torch.cuda.is_available():
            from unsloth.chat_templates import CHAT_TEMPLATES, get_chat_template

            print("\n=== Available Unsloth Chat Templates (CUDA) ===")
        else:
            # Use standalone templates for CPU/MPS
            from autotrain.preprocessor.chat_templates_standalone import CHAT_TEMPLATES

            print("\n=== Available Standalone Chat Templates (CPU/MPS) ===")
            get_chat_template = None

        templates = list(CHAT_TEMPLATES.keys())
        print(f"Found {len(templates)} templates:")

        # Group by family
        families = {}
        for t in templates:
            if "llama" in t.lower():
                family = "Llama"
            elif "gemma" in t.lower():
                family = "Gemma"
            elif "phi" in t.lower():
                family = "Phi"
            elif "qwen" in t.lower():
                family = "Qwen"
            elif t in ["mistral", "zephyr", "chatml", "alpaca", "vicuna"]:
                family = "Common"
            else:
                family = "Other"

            if family not in families:
                families[family] = []
            families[family].append(t)

        for family, temps in sorted(families.items()):
            print(
                f"  {family}: {', '.join(sorted(temps)[:5])}" + (f" + {len(temps)-5} more" if len(temps) > 5 else "")
            )

        return True, templates
    except ImportError as e:
        print(f"WARNING: Unsloth not available: {e}")
        return False, []
    except Exception as e:
        print(f"ERROR validating templates: {e}")
        return False, []


class TestRealDatasetConversion:
    """Test dataset conversion with real datasets."""

    @pytest.fixture(autouse=True)
    def setup(self):
        """Setup test environment and validate templates."""
        self.has_templates, self.available_templates = validate_chat_templates()
        if not self.has_templates:
            pytest.skip("No chat templates available, skipping real conversion tests")

    def test_alpaca_format_detection_and_conversion(self):
        """Test Alpaca dataset format detection and conversion."""
        print("\n=== Testing Alpaca Format ===")

        # Load real Alpaca dataset
        dataset = load_dataset("tatsu-lab/alpaca", split="train[:5]")
        print(f"Loaded {len(dataset)} samples")
        print(f"Columns: {dataset.column_names}")

        # Test format detection
        format_type = detect_dataset_format(dataset)
        assert format_type == "alpaca", f"Expected 'alpaca', got '{format_type}'"
        print(f"✓ Correctly detected format: {format_type}")

        # Convert to messages format
        converted = convert_alpaca_to_messages(dataset)

        # Validate conversion
        assert "messages" in converted.column_names
        sample = converted[0]["messages"]
        assert isinstance(sample, list)
        assert len(sample) >= 2  # At least user and assistant
        assert sample[0]["role"] == "user"
        assert sample[-1]["role"] == "assistant"

        print(f"✓ Converted to messages format")
        print(f"  First message: {sample[0]}")
        print(f"  Last message: {sample[-1]}")

        return converted

    def test_sharegpt_format_detection_and_conversion(self):
        """Test ShareGPT dataset format detection and conversion."""
        print("\n=== Testing ShareGPT Format ===")

        # Load real ShareGPT dataset
        dataset = load_dataset("philschmid/guanaco-sharegpt-style", split="train[:5]")
        print(f"Loaded {len(dataset)} samples")
        print(f"Columns: {dataset.column_names}")

        # Test format detection
        format_type = detect_dataset_format(dataset)
        assert format_type == "sharegpt", f"Expected 'sharegpt', got '{format_type}'"
        print(f"✓ Correctly detected format: {format_type}")

        # Test conversion - will use Unsloth on CUDA, fallback on CPU/MPS
        converted = convert_sharegpt_to_messages(dataset)
        print("✓ Converted ShareGPT format")

        # Validate conversion
        if "messages" in converted.column_names:
            messages_col = "messages"
        elif "conversations" in converted.column_names:
            messages_col = "conversations"
        else:
            raise AssertionError(f"No messages column found in {converted.column_names}")

        sample = converted[0][messages_col]
        assert isinstance(sample, list)
        assert len(sample) >= 1

        # Check role mapping
        first_msg = sample[0]
        assert "role" in first_msg or "from" in first_msg
        if "role" in first_msg:
            assert first_msg["role"] in ["user", "assistant", "system"]

        print(f"✓ Converted ShareGPT format")
        print(f"  Sample conversation length: {len(sample)}")

        return converted

    def test_dpo_format_detection_and_handling(self):
        """Test DPO dataset format detection."""
        print("\n=== Testing DPO Format ===")

        # Load real DPO dataset
        dataset = load_dataset("argilla/ultrafeedback-binarized-preferences-cleaned", split="train[:5]")
        print(f"Loaded {len(dataset)} samples")
        print(f"Columns: {dataset.column_names}")

        # Test format detection
        format_type = detect_dataset_format(dataset)
        assert format_type in ["dpo", "messages"], f"Expected 'dpo' or 'messages', got '{format_type}'"
        print(f"✓ Detected format: {format_type}")

        # DPO datasets often have prompt/chosen/rejected or messages format
        if "prompt" in dataset.column_names and "chosen" in dataset.column_names:
            print("✓ Has DPO columns: prompt, chosen, rejected")

            # Convert if needed
            if format_type == "dpo":
                converted = convert_dpo_to_messages(dataset)
                assert "messages_chosen" in converted.column_names
                assert "messages_rejected" in converted.column_names
                print("✓ Converted DPO to messages format")

        return dataset

    def test_chat_template_application_with_models(self):
        """Test applying chat templates with real models and tokenizers."""
        print("\n=== Testing Chat Template Application ===")

        # Load a small dataset
        dataset = load_dataset("tatsu-lab/alpaca", split="train[:3]")

        # Convert to messages format first
        dataset = convert_alpaca_to_messages(dataset)

        for model_name in TEST_MODELS[:2]:  # Test first 2 models to save time
            print(f"\n--- Testing with {model_name} ---")

            try:
                # Load tokenizer
                tokenizer = AutoTokenizer.from_pretrained(
                    model_name, trust_remote_code=True, token=os.environ.get("HF_TOKEN")
                )

                # Test 1: Use tokenizer's default template
                print("  Testing tokenizer default template...")
                result_default = apply_chat_template(dataset, tokenizer, chat_template=None)  # Use tokenizer default

                assert "text" in result_default.column_names
                sample_text = result_default[0]["text"]
                assert len(sample_text) > 0
                print(f"  ✓ Default template applied, length: {len(sample_text)}")
                print(f"    Preview: {sample_text[:100]}...")

                # Test 2: Try specific template if available
                if self.has_templates:
                    # Determine appropriate template
                    if "llama" in model_name.lower():
                        template = "llama" if "llama-2" in model_name.lower() else "llama3"
                    elif "gemma" in model_name.lower():
                        template = "gemma"
                    elif "phi" in model_name.lower():
                        template = "phi-3"
                    elif "zephyr" in model_name.lower():
                        template = "zephyr"
                    else:
                        template = "chatml"  # Generic fallback

                    if template in self.available_templates:
                        print(f"  Testing template: {template}...")
                        result_specific = apply_chat_template(dataset, tokenizer, chat_template=template)

                        assert "text" in result_specific.column_names
                        sample_text = result_specific[0]["text"]
                        assert len(sample_text) > 0
                        print(f"  ✓ {template} template applied, length: {len(sample_text)}")
                        print(f"    Preview: {sample_text[:100]}...")

            except Exception as e:
                print(f"  ⚠ Error with {model_name}: {e}")
                continue

    def test_conversation_extension(self):
        """Test extending single-turn Alpaca to multi-turn conversations."""
        print("\n=== Testing Conversation Extension ===")

        # Load Alpaca dataset
        dataset = load_dataset("tatsu-lab/alpaca", split="train[:10]")

        # Test different extension values
        for extension in [1, 2, 3]:
            print(f"\n--- Extension factor: {extension} ---")

            extended = extend_alpaca_conversations(
                dataset, conversation_extension=extension, output_column_name="output"
            )

            if extension == 1:
                # No extension, should be same length
                assert len(extended) == len(dataset)
                print(f"  ✓ No extension (factor=1): {len(extended)} samples")
            else:
                # Should have fewer samples (merged)
                expected_len = len(dataset) // extension + (1 if len(dataset) % extension else 0)
                assert len(extended) <= expected_len + 1  # Allow for rounding
                print(f"  ✓ Extended: {len(dataset)} → {len(extended)} samples")

                # Check that conversations are longer
                if "messages" in extended.column_names:
                    sample = extended[0]["messages"]
                    expected_turns = extension * 2  # Each original has 2 messages
                    assert len(sample) >= min(expected_turns, len(dataset) * 2)
                    print(f"    Conversation length: {len(sample)} messages")

    def test_full_pipeline_with_analyze_and_convert(self):
        """Test the complete analyze_and_convert_dataset function."""
        print("\n=== Testing Full Pipeline ===")

        test_cases = [
            ("tatsu-lab/alpaca", "alpaca"),
            ("philschmid/guanaco-sharegpt-style", "sharegpt"),
        ]

        for dataset_name, expected_format in test_cases:
            print(f"\n--- Testing {dataset_name} ---")

            # Load dataset
            dataset = load_dataset(dataset_name, split="train[:5]")

            # Load a tokenizer
            tokenizer = AutoTokenizer.from_pretrained("HuggingFaceH4/zephyr-7b-beta", trust_remote_code=True)

            # Run full analysis and conversion
            result = analyze_and_convert_dataset(
                dataset, tokenizer=tokenizer, chat_template="chatml", trainer_type="sft", apply_template=True
            )

            # Validate result
            assert "format_detected" in result
            assert result["format_detected"] == expected_format
            # The key might be 'dataset' or 'converted_dataset'
            assert "converted_dataset" in result or "dataset" in result

            converted_ds = result.get("converted_dataset", result.get("dataset"))
            if result.get("conversion_applied"):
                # Should have text column if template was applied
                if "text" in converted_ds.column_names:
                    assert len(converted_ds[0]["text"]) > 0
                    print(f"  ✓ Full pipeline successful")
                    print(f"    Format: {result['format_detected']}")
                    print(f"    Converted: {result.get('conversion_applied', False)}")
                    print(f"    Columns: {converted_ds.column_names}")

    def test_template_validation_against_registry(self):
        """Validate that template names match CHAT_TEMPLATES registry."""
        print("\n=== Testing Template Name Validation ===")

        if not self.has_templates:
            pytest.skip("No templates available")

        # Import appropriate CHAT_TEMPLATES
        try:
            import torch

            if torch.cuda.is_available():
                from unsloth.chat_templates import CHAT_TEMPLATES
            else:
                from autotrain.preprocessor.chat_templates_standalone import CHAT_TEMPLATES
        except ImportError:
            from autotrain.preprocessor.chat_templates_standalone import CHAT_TEMPLATES

        # Test valid template names
        valid_templates = ["llama", "llama3", "gemma", "chatml", "zephyr", "mistral"]

        for template_name in valid_templates:
            if template_name in CHAT_TEMPLATES:
                print(f"  ✓ '{template_name}' is valid in template registry")

                # On CPU/MPS we can't test template application with get_chat_template
                # but we can verify the template exists
                try:
                    import torch

                    if torch.cuda.is_available():
                        # Test that we can actually use it with Unsloth
                        tokenizer = AutoTokenizer.from_pretrained(
                            "HuggingFaceH4/zephyr-7b-beta", trust_remote_code=True
                        )
                        from unsloth.chat_templates import get_chat_template

                        enhanced_tokenizer = get_chat_template(tokenizer, chat_template=template_name)
                        assert enhanced_tokenizer is not None
                        print(f"    ✓ Successfully applied '{template_name}' template")
                    else:
                        # Just verify template string exists
                        template_str = CHAT_TEMPLATES[template_name]
                        assert len(template_str) > 0
                        print(f"    ✓ Template '{template_name}' exists ({len(template_str)} chars)")
                except Exception as e:
                    print(f"    ⚠ Could not validate '{template_name}': {e}")

        # Test invalid template handling
        invalid_template = "not_a_real_template_xyz"
        if invalid_template not in CHAT_TEMPLATES:
            print(f"  ✓ '{invalid_template}' correctly identified as invalid")


def test_dataset_format_detection_accuracy():
    """Test accuracy of format detection across multiple datasets."""
    print("\n=== Testing Format Detection Accuracy ===")

    test_cases = [
        # (dataset_id, split, expected_format)
        ("tatsu-lab/alpaca", "train[:5]", "alpaca"),
        ("yahma/alpaca-cleaned", "train[:5]", "alpaca"),
        ("philschmid/guanaco-sharegpt-style", "train[:5]", "sharegpt"),
        ("HuggingFaceH4/no_robots", "train[:5]", "messages"),
        ("openai/gsm8k", "train[:5]", "unknown"),  # Not a chat dataset
    ]

    results = []
    for dataset_id, split, expected in test_cases:
        try:
            dataset = load_dataset(dataset_id, split=split)
            detected = detect_dataset_format(dataset)

            # Some datasets might be detected differently
            # Allow messages/sharegpt to be interchangeable for some cases
            is_correct = (detected == expected) or (
                expected in ["messages", "sharegpt"] and detected in ["messages", "sharegpt"]
            )

            results.append({"dataset": dataset_id, "expected": expected, "detected": detected, "correct": is_correct})

            status = "✓" if is_correct else "✗"
            print(f"  {status} {dataset_id}: {detected} (expected: {expected})")

        except Exception as e:
            print(f"  ⚠ Could not test {dataset_id}: {e}")
            continue

    # Calculate accuracy
    correct = sum(1 for r in results if r["correct"])
    total = len(results)
    accuracy = (correct / total * 100) if total > 0 else 0

    print(f"\nAccuracy: {correct}/{total} ({accuracy:.1f}%)")
    assert accuracy >= 60, f"Format detection accuracy too low: {accuracy}%"


if __name__ == "__main__":
    # Run validation first
    print("=" * 60)
    print("DATASET CONVERSION REAL-WORLD TESTS")
    print("=" * 60)

    has_templates, templates = validate_chat_templates()

    if not has_templates:
        print("\n⚠️  WARNING: No chat templates available!")
        print("\nSome tests will be skipped.")

    # Run pytest with verbose output
    pytest.main([__file__, "-v", "-s"])



"""
Evaluation Metrics for AutoTrain Advanced
==========================================

Collection of metrics for model evaluation.
"""

from abc import ABC, abstractmethod
from typing import Any, Dict, List, Union

import numpy as np
import torch


# Optional imports with fallback
try:
    from sacrebleu import corpus_bleu

    SACREBLEU_AVAILABLE = True
except ImportError:
    SACREBLEU_AVAILABLE = False

try:
    from rouge_score import rouge_scorer

    ROUGE_AVAILABLE = True
except ImportError:
    ROUGE_AVAILABLE = False

try:
    from bert_score import score as bert_score

    BERTSCORE_AVAILABLE = True
except ImportError:
    BERTSCORE_AVAILABLE = False

try:
    import nltk
    from nltk.translate.meteor_score import meteor_score

    METEOR_AVAILABLE = True
except ImportError:
    METEOR_AVAILABLE = False

from sklearn.metrics import accuracy_score, f1_score, precision_score, recall_score


class Metric(ABC):
    """Abstract base class for metrics."""

    @abstractmethod
    def compute(self, predictions: List[Any], references: List[Any], **kwargs) -> Union[float, Dict[str, float]]:
        """Compute metric value."""

    def reset(self):
        """Reset internal state if any."""


class PerplexityMetric(Metric):
    """Perplexity metric for language modeling."""

    def compute(self, losses: Union[List[float], torch.Tensor], **kwargs) -> float:
        """
        Compute perplexity from losses.

        Args:
            losses: List of loss values or loss tensor

        Returns:
            Perplexity value
        """
        if isinstance(losses, torch.Tensor):
            avg_loss = losses.mean().item()
        else:
            avg_loss = np.mean(losses)

        perplexity = np.exp(avg_loss) if avg_loss < 100 else float("inf")
        return perplexity


class BLEUMetric(Metric):
    """BLEU score for text generation."""

    def __init__(self, n_gram: int = 4, smooth: bool = True):
        self.n_gram = n_gram
        self.smooth = smooth

    def compute(self, predictions: List[str], references: List[Union[str, List[str]]], **kwargs) -> Dict[str, float]:
        """
        Compute BLEU score.

        Args:
            predictions: Generated texts
            references: Reference texts (can be multiple per prediction)

        Returns:
            Dictionary with BLEU scores
        """
        if not SACREBLEU_AVAILABLE:
            # Fallback to simple implementation
            return self._simple_bleu(predictions, references)

        # Ensure references is list of lists
        if references and isinstance(references[0], str):
            references = [[ref] for ref in references]

        # Compute BLEU
        bleu = corpus_bleu(predictions, references)

        return {
            "bleu": bleu.score / 100.0,  # Normalize to 0-1
            "bleu_1": bleu.precisions[0] / 100.0 if len(bleu.precisions) > 0 else 0.0,
            "bleu_2": bleu.precisions[1] / 100.0 if len(bleu.precisions) > 1 else 0.0,
            "bleu_3": bleu.precisions[2] / 100.0 if len(bleu.precisions) > 2 else 0.0,
            "bleu_4": bleu.precisions[3] / 100.0 if len(bleu.precisions) > 3 else 0.0,
        }

    def _simple_bleu(self, predictions: List[str], references: List[str]) -> Dict[str, float]:
        """Simple BLEU implementation as fallback."""
        # Very basic n-gram overlap
        scores = []
        for pred, ref in zip(predictions, references):
            if isinstance(ref, list):
                ref = ref[0]

            pred_tokens = pred.split()
            ref_tokens = ref.split()

            if not pred_tokens or not ref_tokens:
                scores.append(0.0)
                continue

            # Count matching n-grams
            matches = 0
            for i in range(len(pred_tokens)):
                if i < len(ref_tokens) and pred_tokens[i] == ref_tokens[i]:
                    matches += 1

            score = matches / max(len(pred_tokens), 1)
            scores.append(score)

        return {"bleu": np.mean(scores)}


class ROUGEMetric(Metric):
    """ROUGE scores for text generation."""

    def __init__(self, rouge_types: List[str] = None):
        self.rouge_types = rouge_types or ["rouge1", "rouge2", "rougeL"]

    def compute(self, predictions: List[str], references: List[str], **kwargs) -> Dict[str, float]:
        """
        Compute ROUGE scores.

        Args:
            predictions: Generated texts
            references: Reference texts

        Returns:
            Dictionary with ROUGE scores
        """
        if not ROUGE_AVAILABLE:
            # Fallback to simple implementation
            return self._simple_rouge(predictions, references)

        scorer = rouge_scorer.RougeScorer(self.rouge_types, use_stemmer=True)
        scores = {rouge_type: [] for rouge_type in self.rouge_types}

        for pred, ref in zip(predictions, references):
            if isinstance(ref, list):
                ref = ref[0]

            result = scorer.score(ref, pred)
            for rouge_type in self.rouge_types:
                scores[rouge_type].append(result[rouge_type].fmeasure)

        # Average scores
        avg_scores = {}
        for rouge_type in self.rouge_types:
            avg_scores[rouge_type] = np.mean(scores[rouge_type])

        return avg_scores

    def _simple_rouge(self, predictions: List[str], references: List[str]) -> Dict[str, float]:
        """Simple ROUGE implementation as fallback."""
        scores = []
        for pred, ref in zip(predictions, references):
            if isinstance(ref, list):
                ref = ref[0]

            pred_tokens = set(pred.split())
            ref_tokens = set(ref.split())

            if not pred_tokens or not ref_tokens:
                scores.append(0.0)
                continue

            overlap = len(pred_tokens & ref_tokens)
            precision = overlap / len(pred_tokens) if pred_tokens else 0
            recall = overlap / len(ref_tokens) if ref_tokens else 0
            f1 = 2 * precision * recall / (precision + recall) if (precision + recall) > 0 else 0

            scores.append(f1)

        return {"rouge1": np.mean(scores)}


class BERTScoreMetric(Metric):
    """BERTScore for semantic similarity."""

    def __init__(self, model_name: str = "bert-base-uncased", device: str = "auto"):
        self.model_name = model_name

        # Auto-detect device
        if device == "auto":
            if torch.cuda.is_available():
                self.device = "cuda"
            elif torch.backends.mps.is_available():
                self.device = "mps"
            else:
                self.device = "cpu"
        else:
            self.device = device

    def compute(self, predictions: List[str], references: List[str], **kwargs) -> Dict[str, float]:
        """
        Compute BERTScore.

        Args:
            predictions: Generated texts
            references: Reference texts

        Returns:
            Dictionary with BERTScore values
        """
        if not BERTSCORE_AVAILABLE:
            return {"bert_score_f1": 0.0}

        # Compute BERTScore
        P, R, F1 = bert_score(predictions, references, model_type=self.model_name, device=self.device, verbose=False)

        return {
            "bert_score_precision": P.mean().item(),
            "bert_score_recall": R.mean().item(),
            "bert_score_f1": F1.mean().item(),
        }


class AccuracyMetric(Metric):
    """Accuracy metric for classification."""

    def compute(
        self, predictions: Union[List[int], np.ndarray], references: Union[List[int], np.ndarray], **kwargs
    ) -> float:
        """
        Compute accuracy.

        Args:
            predictions: Predicted labels
            references: True labels

        Returns:
            Accuracy value
        """
        return accuracy_score(references, predictions)


class F1Metric(Metric):
    """F1 score for classification."""

    def __init__(self, average: str = "macro"):
        self.average = average

    def compute(
        self, predictions: Union[List[int], np.ndarray], references: Union[List[int], np.ndarray], **kwargs
    ) -> Dict[str, float]:
        """
        Compute F1 score.

        Args:
            predictions: Predicted labels
            references: True labels

        Returns:
            Dictionary with F1, precision, recall
        """
        f1 = f1_score(references, predictions, average=self.average, zero_division=0)
        precision = precision_score(references, predictions, average=self.average, zero_division=0)
        recall = recall_score(references, predictions, average=self.average, zero_division=0)

        return {
            "f1": f1,
            "precision": precision,
            "recall": recall,
        }


class ExactMatchMetric(Metric):
    """Exact match metric for QA and generation."""

    def __init__(self, normalize: bool = True, ignore_case: bool = True):
        self.normalize = normalize
        self.ignore_case = ignore_case

    def compute(self, predictions: List[str], references: List[str], **kwargs) -> float:
        """
        Compute exact match score.

        Args:
            predictions: Predicted texts
            references: Reference texts

        Returns:
            Exact match score
        """
        matches = []
        for pred, ref in zip(predictions, references):
            if isinstance(ref, list):
                ref = ref[0]

            if self.normalize:
                pred = self._normalize_text(pred)
                ref = self._normalize_text(ref)

            if self.ignore_case:
                pred = pred.lower()
                ref = ref.lower()

            matches.append(float(pred == ref))

        return np.mean(matches)

    def _normalize_text(self, text: str) -> str:
        """Normalize text for comparison."""
        import re

        # Remove extra whitespace
        text = " ".join(text.split())
        # Remove punctuation
        text = re.sub(r"[^\w\s]", "", text)
        return text.strip()


class METEORMetric(Metric):
    """METEOR score for text generation."""

    def compute(self, predictions: List[str], references: List[str], **kwargs) -> float:
        """
        Compute METEOR score.

        Args:
            predictions: Generated texts
            references: Reference texts

        Returns:
            METEOR score
        """
        if not METEOR_AVAILABLE:
            # Fallback to simple overlap
            scores = []
            for pred, ref in zip(predictions, references):
                if isinstance(ref, list):
                    ref = ref[0]

                pred_tokens = set(pred.split())
                ref_tokens = set(ref.split())

                if not pred_tokens or not ref_tokens:
                    scores.append(0.0)
                    continue

                overlap = len(pred_tokens & ref_tokens)
                score = overlap / max(len(pred_tokens), len(ref_tokens))
                scores.append(score)

            return np.mean(scores)

        # Download required NLTK data
        try:
            nltk.download("wordnet", quiet=True)
            nltk.download("punkt", quiet=True)
            nltk.download("omw-1.4", quiet=True)
        except:
            pass

        scores = []
        for pred, ref in zip(predictions, references):
            if isinstance(ref, list):
                ref = ref[0]

            try:
                score = meteor_score([ref], pred)
                scores.append(score)
            except:
                scores.append(0.0)

        return np.mean(scores)


class CustomMetric(Metric):
    """Custom metric with user-defined function."""

    def __init__(self, metric_fn: callable, name: str = "custom"):
        """
        Initialize custom metric.

        Args:
            metric_fn: Function that takes (predictions, references) and returns score
            name: Name of the metric
        """
        self.metric_fn = metric_fn
        self.name = name

    def compute(self, predictions: List[Any], references: List[Any], **kwargs) -> Union[float, Dict[str, float]]:
        """Compute custom metric."""
        return self.metric_fn(predictions, references, **kwargs)


class MetricCollection:
    """Collection of metrics to compute together."""

    def __init__(self, metrics: Dict[str, Metric]):
        """
        Initialize metric collection.

        Args:
            metrics: Dictionary of metric name to Metric instance
        """
        self.metrics = metrics

    def compute_all(
        self, predictions: List[Any], references: List[Any], **kwargs
    ) -> Dict[str, Union[float, Dict[str, float]]]:
        """
        Compute all metrics.

        Args:
            predictions: Model predictions
            references: Reference values

        Returns:
            Dictionary of metric results
        """
        results = {}
        for name, metric in self.metrics.items():
            try:
                result = metric.compute(predictions, references, **kwargs)
                if isinstance(result, dict):
                    # Flatten nested metrics
                    for sub_name, value in result.items():
                        results[f"{name}_{sub_name}" if sub_name != name else name] = value
                else:
                    results[name] = result
            except Exception as e:
                print(f"Failed to compute {name}: {e}")
                results[name] = None

        return results

    def add_metric(self, name: str, metric: Metric):
        """Add a metric to the collection."""
        self.metrics[name] = metric

    def remove_metric(self, name: str):
        """Remove a metric from the collection."""
        if name in self.metrics:
            del self.metrics[name]



"""
Completer implementations for different generation levels
==========================================================

Provides TokenCompleter and MessageCompleter abstractions.
"""

import asyncio
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, AsyncIterator, Dict, Iterator, List, Optional, Union

import torch
from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
    GenerationConfig,
    LogitsProcessor,
    StoppingCriteria,
    StoppingCriteriaList,
)

from autotrain.rendering import ChatFormat, Conversation, Message, RenderConfig, get_renderer


class MultiTokenStoppingCriteria(StoppingCriteria):
    """Stopping criteria that detects multi-token stop sequences during generation."""

    def __init__(self, stop_sequences: List[str], tokenizer: AutoTokenizer, input_length: int):
        """
        Initialize stopping criteria.

        Args:
            stop_sequences: List of strings to stop at (e.g., ['<end_of_turn>', '<|im_end|>'])
            tokenizer: The tokenizer to decode generated tokens
            input_length: Length of input prompt tokens (to skip checking prompt)
        """
        self.stop_sequences = stop_sequences
        self.tokenizer = tokenizer
        self.input_length = input_length

        # Pre-tokenize stop sequences for faster matching
        self.stop_token_ids = []
        for seq in stop_sequences:
            # Get token IDs for each stop sequence
            tokens = tokenizer.encode(seq, add_special_tokens=False)
            self.stop_token_ids.append(tokens)

    def __call__(self, input_ids: torch.LongTensor, scores: torch.FloatTensor, **kwargs) -> bool:
        """
        Check if any stop sequence appears in the generated tokens.

        Args:
            input_ids: Generated token IDs [batch_size, sequence_length]
            scores: Logits scores (unused but required by interface)

        Returns:
            True if generation should stop, False otherwise
        """
        # Only check the newly generated tokens (skip input prompt)
        generated_ids = input_ids[0, self.input_length :]

        # Check each stop sequence
        for stop_tokens in self.stop_token_ids:
            stop_len = len(stop_tokens)

            # Need at least enough tokens to match
            if len(generated_ids) < stop_len:
                continue

            # Check if the last N tokens match this stop sequence
            if torch.equal(generated_ids[-stop_len:], torch.tensor(stop_tokens, device=generated_ids.device)):
                return True

        return False


@dataclass
class CompletionConfig:
    """Configuration for text completion."""

    # Generation parameters
    max_new_tokens: int = 256
    min_new_tokens: int = 1
    temperature: float = 1.0
    top_p: float = 1.0
    top_k: int = 50
    repetition_penalty: float = 1.0
    length_penalty: float = 1.0

    # Sampling
    do_sample: bool = True
    num_beams: int = 1
    num_return_sequences: int = 1

    # Early stopping
    early_stopping: bool = False
    max_time: Optional[float] = None

    # Token control
    eos_token_id: Optional[Union[int, List[int]]] = None
    pad_token_id: Optional[int] = None
    bos_token_id: Optional[int] = None

    # Advanced
    use_cache: bool = True
    return_dict_in_generate: bool = True
    output_scores: bool = False
    output_logits: bool = False
    output_attentions: bool = False

    # Chat format (for MessageCompleter)
    chat_format: ChatFormat = ChatFormat.CHATML
    add_generation_prompt: bool = True

    # Custom processors
    logits_processor: Optional[LogitsProcessor] = None
    stopping_criteria: Optional[StoppingCriteriaList] = None

    def to_generation_config(self) -> GenerationConfig:
        """Convert to HuggingFace GenerationConfig."""
        return GenerationConfig(
            max_new_tokens=self.max_new_tokens,
            min_new_tokens=self.min_new_tokens,
            temperature=self.temperature,
            top_p=self.top_p,
            top_k=self.top_k,
            repetition_penalty=self.repetition_penalty,
            length_penalty=self.length_penalty,
            do_sample=self.do_sample,
            num_beams=self.num_beams,
            num_return_sequences=self.num_return_sequences,
            early_stopping=self.early_stopping,
            max_time=self.max_time,
            eos_token_id=self.eos_token_id,
            pad_token_id=self.pad_token_id,
            bos_token_id=self.bos_token_id,
            use_cache=self.use_cache,
            return_dict_in_generate=self.return_dict_in_generate,
            output_scores=self.output_scores,
            output_attentions=self.output_attentions,
        )


@dataclass
class CompletionResult:
    """Base result class for completions."""

    text: str
    tokens: Optional[torch.Tensor] = None
    scores: Optional[torch.Tensor] = None
    logits: Optional[torch.Tensor] = None
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class TokenCompletionResult(CompletionResult):
    """Result from token-level completion."""

    token_ids: torch.Tensor = None
    token_scores: Optional[torch.Tensor] = None
    token_logprobs: Optional[torch.Tensor] = None

    @property
    def num_tokens(self) -> int:
        """Get number of generated tokens."""
        if self.token_ids is not None:
            return self.token_ids.shape[-1]
        return 0


@dataclass
class MessageCompletionResult(CompletionResult):
    """Result from message-level completion."""

    message: Message = None
    conversation: Optional[Conversation] = None
    stop_reason: Optional[str] = None

    @property
    def content(self) -> str:
        """Get message content."""
        return self.message.content if self.message else self.text


class Completer(ABC):
    """Abstract base class for text completers."""

    def __init__(
        self,
        model: AutoModelForCausalLM,
        tokenizer: AutoTokenizer,
        config: Optional[CompletionConfig] = None,
    ):
        self.model = model
        self.tokenizer = tokenizer
        self.config = config or CompletionConfig()

        # Set special tokens if not configured
        if self.config.eos_token_id is None:
            self.config.eos_token_id = tokenizer.eos_token_id
        if self.config.pad_token_id is None:
            self.config.pad_token_id = tokenizer.pad_token_id or tokenizer.eos_token_id
        if self.config.bos_token_id is None:
            self.config.bos_token_id = tokenizer.bos_token_id

    @abstractmethod
    def complete(self, prompt: Any, **kwargs) -> CompletionResult:
        """Generate completion for prompt."""

    @abstractmethod
    def batch_complete(self, prompts: List[Any], **kwargs) -> List[CompletionResult]:
        """Generate completions for multiple prompts."""


class TokenCompleter(Completer):
    """Low-level token generation for RL and fine-grained control."""

    def complete(
        self,
        prompt: Union[str, torch.Tensor],
        max_new_tokens: Optional[int] = None,
        temperature: Optional[float] = None,
        return_logprobs: bool = False,
        **kwargs,
    ) -> TokenCompletionResult:
        """
        Generate tokens from prompt.

        Args:
            prompt: String or token tensor
            max_new_tokens: Override config setting
            temperature: Override config temperature
            return_logprobs: Whether to return log probabilities

        Returns:
            TokenCompletionResult with generated tokens
        """
        # Convert string to tokens if needed
        if isinstance(prompt, str):
            input_ids = self.tokenizer.encode(prompt, return_tensors="pt")
        else:
            input_ids = prompt

        # Move to model device
        input_ids = input_ids.to(self.model.device)

        # Update generation config
        gen_config = self.config.to_generation_config()
        if max_new_tokens is not None:
            gen_config.max_new_tokens = max_new_tokens
        if temperature is not None:
            gen_config.temperature = temperature

        # Generate
        with torch.no_grad():
            outputs = self.model.generate(
                input_ids,
                generation_config=gen_config,
                logits_processor=self.config.logits_processor,
                stopping_criteria=self.config.stopping_criteria,
                **kwargs,
            )

        # Extract generated tokens (remove prompt)
        generated_ids = outputs.sequences if hasattr(outputs, "sequences") else outputs
        generated_ids = generated_ids[:, input_ids.shape[1] :]

        # Decode text
        generated_text = self.tokenizer.decode(generated_ids[0], skip_special_tokens=True)

        # Compute log probabilities if requested
        token_logprobs = None
        if return_logprobs and hasattr(outputs, "scores"):
            # Stack scores and compute log probs
            scores = torch.stack(outputs.scores, dim=1)
            token_logprobs = torch.log_softmax(scores, dim=-1)

            # Get log probs for generated tokens
            batch_size = generated_ids.shape[0]
            token_logprobs_selected = []
            for b in range(batch_size):
                logprobs_b = []
                for t in range(generated_ids.shape[1]):
                    token_id = generated_ids[b, t]
                    logprob = token_logprobs[b, t, token_id]
                    logprobs_b.append(logprob)
                token_logprobs_selected.append(torch.stack(logprobs_b))
            token_logprobs = torch.stack(token_logprobs_selected)

        return TokenCompletionResult(
            text=generated_text,
            tokens=generated_ids,
            token_ids=generated_ids[0],
            token_scores=outputs.scores if hasattr(outputs, "scores") else None,
            token_logprobs=token_logprobs,
            logits=outputs.logits if hasattr(outputs, "logits") else None,
            metadata={
                "num_tokens": generated_ids.shape[1],
                "model": self.model.config.name_or_path,
            },
        )

    def batch_complete(self, prompts: List[Union[str, torch.Tensor]], **kwargs) -> List[TokenCompletionResult]:
        """Generate completions for multiple prompts."""
        results = []

        # Process each prompt
        # TODO: Implement true batch processing for efficiency
        for prompt in prompts:
            result = self.complete(prompt, **kwargs)
            results.append(result)

        return results

    def stream_tokens(self, prompt: Union[str, torch.Tensor], **kwargs) -> Iterator[torch.Tensor]:
        """
        Stream tokens as they are generated.

        Yields:
            Individual token IDs
        """
        # This is a simplified version - real streaming would use
        # a custom generation loop
        result = self.complete(prompt, **kwargs)

        for i in range(result.num_tokens):
            yield result.token_ids[i]


class MessageCompleter(Completer):
    """High-level message generation with chat formatting."""

    def __init__(
        self,
        model: AutoModelForCausalLM,
        tokenizer: AutoTokenizer,
        config: Optional[CompletionConfig] = None,
        chat_format: Optional[ChatFormat] = None,
    ):
        super().__init__(model, tokenizer, config)

        # Set chat format
        if chat_format:
            self.config.chat_format = chat_format

        # Create renderer
        render_config = RenderConfig(
            format=self.config.chat_format,
            add_generation_prompt=self.config.add_generation_prompt,
        )
        self.renderer = get_renderer(self.config.chat_format, tokenizer, render_config)

        # Get stop sequences
        self.stop_sequences = self.renderer.get_stop_sequences()

    def complete(
        self, conversation: Union[Conversation, List[Dict[str, str]]], system_prompt: Optional[str] = None, **kwargs
    ) -> MessageCompletionResult:
        """
        Generate a message response.

        Args:
            conversation: Conversation or list of message dicts
            system_prompt: Optional system prompt to prepend

        Returns:
            MessageCompletionResult with generated message
        """
        # Convert to Conversation if needed
        if isinstance(conversation, list):
            messages = []
            if system_prompt:
                messages.append(Message(role="system", content=system_prompt))
            messages.extend(
                [
                    Message(
                        role=m["role"],
                        content=m.get("content") or "",
                        tool_calls=m.get("tool_calls"),
                        tool_call_id=m.get("tool_call_id"),
                    )
                    for m in conversation
                ]
            )
            conversation = Conversation(messages=messages)
        elif system_prompt:
            # Prepend system message
            messages = [Message(role="system", content=system_prompt)] + conversation.messages
            conversation = Conversation(messages=messages)

        # Build generation prompt
        prompt = self.renderer.build_generation_prompt(conversation)

        # Tokenize
        input_ids = self.tokenizer.encode(prompt, return_tensors="pt").to(self.model.device)

        # Generate
        gen_config = self.config.to_generation_config()

        # Handle stop sequences with proper early stopping
        eos_token_id = self.config.eos_token_id
        if eos_token_id is None and hasattr(self.tokenizer, "eos_token_id"):
            eos_token_id = self.tokenizer.eos_token_id

        # Create stopping criteria for multi-token stop sequences
        stopping_criteria = None
        if self.stop_sequences:
            stopping_criteria = StoppingCriteriaList(
                [
                    MultiTokenStoppingCriteria(
                        stop_sequences=self.stop_sequences, tokenizer=self.tokenizer, input_length=input_ids.shape[1]
                    )
                ]
            )

        with torch.no_grad():
            outputs = self.model.generate(
                input_ids,
                generation_config=gen_config,
                eos_token_id=eos_token_id,
                stopping_criteria=stopping_criteria,
                **kwargs,
            )

        # Extract generated text
        generated_ids = outputs.sequences if hasattr(outputs, "sequences") else outputs
        generated_ids = generated_ids[:, input_ids.shape[1] :]
        generated_text = self.tokenizer.decode(generated_ids[0], skip_special_tokens=False)

        # Truncate at stop sequences
        stop_reason = "eos"
        truncated_text = generated_text
        if self.stop_sequences:
            # Find the earliest stop sequence
            earliest_pos = len(generated_text)
            earliest_stop = None
            for stop_seq in self.stop_sequences:
                pos = generated_text.find(stop_seq)
                if pos != -1 and pos < earliest_pos:
                    earliest_pos = pos
                    earliest_stop = stop_seq

            if earliest_stop:
                truncated_text = generated_text[:earliest_pos]
                stop_reason = f"stop_sequence:{earliest_stop}"

        # Parse response
        parsed_content = self.renderer.parse_response(truncated_text)

        # Create message
        generated_message = Message(
            role="assistant",
            content=parsed_content,
            metadata={
                "raw_text": truncated_text,
                "full_generated_text": generated_text,  # Keep original for debugging
                "num_tokens": generated_ids.shape[1],
                "stop_reason": stop_reason,
            },
        )

        # Update conversation
        updated_conversation = Conversation(
            messages=conversation.messages + [generated_message], metadata=conversation.metadata
        )

        return MessageCompletionResult(
            text=parsed_content,
            message=generated_message,
            conversation=updated_conversation,
            stop_reason=stop_reason,
            tokens=generated_ids,
            metadata={
                "model": self.model.config.name_or_path,
                "chat_format": self.config.chat_format.value,
            },
        )

    def batch_complete(
        self, conversations: List[Union[Conversation, List[Dict[str, str]]]], **kwargs
    ) -> List[MessageCompletionResult]:
        """Generate message completions for multiple conversations."""
        results = []

        for conv in conversations:
            result = self.complete(conv, **kwargs)
            results.append(result)

        return results

    def chat(
        self,
        user_input: str,
        conversation: Optional[Conversation] = None,
        system_prompt: Optional[str] = None,
        **kwargs,
    ) -> MessageCompletionResult:
        """
        Simple chat interface.

        Args:
            user_input: User message
            conversation: Optional existing conversation
            system_prompt: System prompt

        Returns:
            Assistant response
        """
        # Create or update conversation
        if conversation is None:
            messages = []
            if system_prompt:
                messages.append(Message(role="system", content=system_prompt))
            conversation = Conversation(messages=messages)

        # Add user message
        conversation.add_message("user", user_input)

        # Generate response
        return self.complete(conversation, **kwargs)


class AsyncTokenCompleter(TokenCompleter):
    """Async version of TokenCompleter."""

    async def complete_async(self, prompt: Union[str, torch.Tensor], **kwargs) -> TokenCompletionResult:
        """Async token generation."""
        # Run generation in thread pool
        loop = asyncio.get_event_loop()
        result = await loop.run_in_executor(None, self.complete, prompt, **kwargs)
        return result

    async def stream_tokens_async(self, prompt: Union[str, torch.Tensor], **kwargs) -> AsyncIterator[torch.Tensor]:
        """Async streaming token generation."""
        result = await self.complete_async(prompt, **kwargs)

        for i in range(result.num_tokens):
            yield result.token_ids[i]
            await asyncio.sleep(0)  # Allow other tasks


class AsyncMessageCompleter(MessageCompleter):
    """Async version of MessageCompleter."""

    async def complete_async(
        self, conversation: Union[Conversation, List[Dict[str, str]]], **kwargs
    ) -> MessageCompletionResult:
        """Async message generation."""
        loop = asyncio.get_event_loop()
        result = await loop.run_in_executor(None, self.complete, conversation, **kwargs)
        return result

    async def chat_async(
        self, user_input: str, conversation: Optional[Conversation] = None, **kwargs
    ) -> MessageCompletionResult:
        """Async chat interface."""
        if conversation is None:
            conversation = Conversation(messages=[])

        conversation.add_message("user", user_input)
        return await self.complete_async(conversation, **kwargs)



"""
Generation/Completers Interface for AutoTrain Advanced
========================================================

Provides clean abstractions for model generation at different levels:
- TokenCompleter: Low-level token generation (for RL)
- MessageCompleter: High-level message generation
- AsyncCompleter: Asynchronous generation support
"""

from .completers import (
    AsyncMessageCompleter,
    AsyncTokenCompleter,
    Completer,
    CompletionConfig,
    CompletionResult,
    MessageCompleter,
    MessageCompletionResult,
    TokenCompleter,
    TokenCompletionResult,
)
from .sampling import BeamSearchSampler, SamplingConfig, SamplingStrategy, TopKSampler, TopPSampler, TypicalSampler
from .utils import ChatSession, batch_complete, create_chat_session, create_completer, stream_messages, stream_tokens


__all__ = [
    # Core completers
    "Completer",
    "TokenCompleter",
    "MessageCompleter",
    "AsyncTokenCompleter",
    "AsyncMessageCompleter",
    # Configurations
    "CompletionConfig",
    "CompletionResult",
    "TokenCompletionResult",
    "MessageCompletionResult",
    # Sampling
    "SamplingConfig",
    "SamplingStrategy",
    "TopKSampler",
    "TopPSampler",
    "BeamSearchSampler",
    "TypicalSampler",
    # Utils
    "create_completer",
    "batch_complete",
    "stream_tokens",
    "stream_messages",
    "create_chat_session",
    "ChatSession",
]



"""Widgets for displaying curated model/dataset catalogs inside the TUI."""

from __future__ import annotations

import logging
from typing import List, Optional

from rich.text import Text
from textual.app import ComposeResult
from textual.containers import Container, Vertical
from textual.message import Message
from textual.widgets import Label, ListItem, ListView, Static

from autotrain.metadata.catalog import CatalogEntry, get_popular_datasets, get_popular_models


logger = logging.getLogger(__name__)


class CatalogPanel(Static):
    """Enhanced catalog view with improved visual hierarchy and feedback.

    Features:
    - Bold labels with faint descriptions
    - Subtle toast/log notifications on selection
    - Better visual separation between models and datasets
    """

    DEFAULT_CSS = """
    CatalogPanel {
        height: 100%;
        padding: 0;
    }

    #catalog-container {
        height: 100%;
        padding: 0;
    }

    .catalog-section {
        height: 50%;
        padding: 1;
        border: round $primary;
        margin: 0 0 1 0;
        background: $panel;
    }

    .catalog-title {
        text-style: bold;
        color: $primary;
        padding: 1;
        background: $primary 10%;
        border-bottom: solid $primary;
        margin: 0 0 1 0;
    }

    #catalog-models-list {
        height: 1fr;
        background: $panel;
        border: none;
    }

    #catalog-datasets-list {
        height: 1fr;
        background: $panel;
        border: none;
    }

    .catalog-item-label {
        text-style: bold;
        color: $text;
    }

    .catalog-item-desc {
        color: $text-muted;
        text-style: italic;
    }

    .empty-state {
        text-align: center;
        color: $text-muted;
        padding: 2;
        text-style: italic;
    }
    """

    class ApplyModel(Message):
        """Posted when the user selects a model."""

        def __init__(self, model_id: str):
            self.model_id = model_id
            super().__init__()

    class ApplyDataset(Message):
        """Posted when the user selects a dataset."""

        def __init__(self, dataset_id: str):
            self.dataset_id = dataset_id
            super().__init__()

    class ShowToast(Message):
        """Posted to request a toast notification."""

        def __init__(self, message: str, type: str = "success"):
            self.message = message
            self.type = type
            super().__init__()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._model_entries: List[CatalogEntry] = []
        self._dataset_entries: List[CatalogEntry] = []

    def compose(self) -> ComposeResult:
        with Vertical(id="catalog-container"):
            # Models section
            with Container(classes="catalog-section"):
                yield Static("✨ Popular Models", classes="catalog-title")
                yield ListView(id="catalog-models-list")

            # Datasets section
            with Container(classes="catalog-section"):
                yield Static("📚 Popular Datasets", classes="catalog-title")
                yield ListView(id="catalog-datasets-list")

    def update_catalog(self, trainer_type: str = "llm", trainer_variant: Optional[str] = None) -> None:
        """Refresh catalog entries based on the selected trainer."""
        try:
            self._model_entries = get_popular_models(trainer_type, trainer_variant)[:12]
            self._dataset_entries = get_popular_datasets(trainer_type, trainer_variant)[:12]
            self._refresh_list("catalog-models-list", self._model_entries, empty_message="No curated models available")
            self._refresh_list(
                "catalog-datasets-list", self._dataset_entries, empty_message="No curated datasets available"
            )
            logger.debug(
                f"Catalog updated for {trainer_type}/{trainer_variant}: {len(self._model_entries)} models, {len(self._dataset_entries)} datasets"
            )
        except Exception as e:
            logger.error(f"Failed to update catalog: {e}")

    def _refresh_list(self, list_id: str, entries: List[CatalogEntry], empty_message: str) -> None:
        """Refresh a catalog list with enhanced formatting."""
        try:
            list_view = self.query_one(f"#{list_id}", ListView)
        except Exception:
            # Not mounted yet
            return

        list_view.clear()

        if not entries:
            list_view.append(ListItem(Label(empty_message, classes="empty-state")))
            return

        for entry in entries:
            # Create rich text with bold label and faint description
            label_text = Text()

            # Bold label
            label = entry.label if entry.label else entry.id
            label_text.append(label, style="bold")

            # Faint description
            if entry.description:
                label_text.append("\n  ")
                label_text.append(entry.description, style="dim italic")

            item = CatalogListItem(entry, label_text)
            list_view.append(item)

    def on_mount(self) -> None:
        """Initialize the catalog with default entries when mounted."""
        # Populate with default LLM catalog entries
        self.update_catalog("llm", "default")

    def on_list_view_selected(self, event: ListView.Selected) -> None:
        """Dispatch apply events with toast notification when an item is selected."""
        item = event.item
        if not isinstance(item, CatalogListItem):
            return

        # Determine which list and post appropriate message
        if event.list_view.id == "catalog-models-list":
            model_name = item.entry.label if item.entry.label else item.entry.id
            self.post_message(self.ApplyModel(item.entry.id))
            # Post toast notification
            self.post_message(self.ShowToast(f"✓ Applied model: {model_name}", "success"))
        elif event.list_view.id == "catalog-datasets-list":
            dataset_name = item.entry.label if item.entry.label else item.entry.id
            self.post_message(self.ApplyDataset(item.entry.id))
            # Post toast notification
            self.post_message(self.ShowToast(f"✓ Applied dataset: {dataset_name}", "success"))


class CatalogListItem(ListItem):
    """List item that stores the associated catalog entry with enhanced formatting."""

    def __init__(self, entry: CatalogEntry, label_text: Text | str):
        if isinstance(label_text, str):
            label_text = Text(label_text)
        super().__init__(Label(label_text))
        self.entry = entry



"""
Utility functions for message rendering
========================================

Helper functions for common rendering tasks.
"""

from typing import Any, Dict, List, Optional, Union

import torch
from transformers import AutoTokenizer

from .message_renderer import ChatFormat, Conversation, Message, RenderConfig, TokenWeight, get_renderer


# Cache for tool role support detection per tokenizer
_tool_role_support_cache: Dict[str, bool] = {}

# Cache for tool_calls field support detection per tokenizer
_tool_calls_support_cache: Dict[str, bool] = {}


def _get_tokenizer_id(tokenizer: AutoTokenizer) -> str:
    """Get a unique identifier for a tokenizer for caching."""
    if hasattr(tokenizer, "name_or_path"):
        return tokenizer.name_or_path
    return str(id(tokenizer))


def check_tool_role_support(tokenizer: AutoTokenizer) -> bool:
    """Check if a tokenizer's chat template supports the 'tool' role.

    Tests by attempting to render a minimal conversation with a tool message.
    The result is cached per tokenizer for performance.

    Args:
        tokenizer: The tokenizer to check

    Returns:
        True if tokenizer supports tool role, False otherwise
    """
    tokenizer_id = _get_tokenizer_id(tokenizer)

    if tokenizer_id in _tool_role_support_cache:
        return _tool_role_support_cache[tokenizer_id]

    # Test with a minimal tool conversation
    test_messages = [
        {"role": "user", "content": "test"},
        {"role": "assistant", "content": "test"},
        {"role": "tool", "content": "test"},
        {"role": "assistant", "content": "test"},
    ]

    try:
        tokenizer.apply_chat_template(test_messages, tokenize=False)
        result = True
    except Exception:
        result = False

    _tool_role_support_cache[tokenizer_id] = result
    return result


def check_tool_calls_support(tokenizer: AutoTokenizer) -> bool:
    """Check if a tokenizer's chat template supports the 'tool_calls' field.

    Tests by attempting to render a minimal conversation with an assistant message
    that has tool_calls. The result is cached per tokenizer for performance.

    Args:
        tokenizer: The tokenizer to check

    Returns:
        True if tokenizer supports tool_calls field, False otherwise
    """
    tokenizer_id = _get_tokenizer_id(tokenizer)

    if tokenizer_id in _tool_calls_support_cache:
        return _tool_calls_support_cache[tokenizer_id]

    # Test with a minimal tool_calls message
    test_messages = [
        {"role": "user", "content": "test"},
        {
            "role": "assistant",
            "content": None,  # Often None when tool_calls is present
            "tool_calls": [
                {
                    "id": "call_123",
                    "type": "function",
                    "function": {"name": "test_func", "arguments": "{}"},
                }
            ],
        },
    ]

    try:
        result_text = tokenizer.apply_chat_template(test_messages, tokenize=False)
        # Check if the tool call information appears in output (model handled it)
        # If it just silently dropped it, we should serialize instead
        if "test_func" in result_text or "call_123" in result_text:
            result = True
        else:
            # Tokenizer didn't include tool_calls in output - doesn't support it
            result = False
    except Exception:
        result = False

    _tool_calls_support_cache[tokenizer_id] = result
    return result


def serialize_tool_calls_to_content(messages: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """Serialize tool_calls field into message content for models that don't support it natively.

    This function:
    1. Finds assistant messages with tool_calls field
    2. Appends the tool calls as JSON to the message content
    3. Returns messages without the tool_calls field (content contains the JSON)

    Args:
        messages: List of message dicts, potentially with tool_calls

    Returns:
        Messages with tool_calls serialized into content

    Example:
        Input:
        {
            "role": "assistant",
            "content": "Let me check that for you.",
            "tool_calls": [{"function": {"name": "search", "arguments": "{\"query\": \"weather\"}"}}]
        }

        Output:
        {
            "role": "assistant",
            "content": "Let me check that for you.\n[Tool Call] {\"tool\": \"search\", \"arguments\": {\"query\": \"weather\"}}"
        }
    """
    import json

    processed = []
    for msg in messages:
        msg_copy = dict(msg)
        tool_calls = msg_copy.pop("tool_calls", None)

        if tool_calls and msg_copy.get("role") == "assistant":
            content = msg_copy.get("content") or ""

            # Serialize each tool call
            for tc in tool_calls:
                func = tc.get("function", {})
                tool_name = func.get("name", "unknown")

                # Parse arguments if they're a string
                args = func.get("arguments", "{}")
                if isinstance(args, str):
                    try:
                        args = json.loads(args)
                    except json.JSONDecodeError:
                        pass  # Keep as string if can't parse

                tool_json = json.dumps({"tool": tool_name, "arguments": args}, ensure_ascii=False)
                content = f"{content}\n[Tool Call] {tool_json}" if content else f"[Tool Call] {tool_json}"

            msg_copy["content"] = content.strip()

        # Remove tool_call_id from tool response messages (handled separately by tool role conversion)
        msg_copy.pop("tool_call_id", None)

        processed.append(msg_copy)

    return processed


def preprocess_messages_for_tool_role(messages: List[Dict[str, str]]) -> List[Dict[str, str]]:
    """Preprocess messages to handle 'tool' role for tokenizers that don't support it.

    This function:
    1. Converts 'tool' role to 'user' role with [Tool Result] prefix
    2. Merges consecutive same-role messages to maintain strict alternation

    Args:
        messages: List of message dicts with 'role' and 'content' keys

    Returns:
        Preprocessed messages compatible with strict-alternation tokenizers
    """
    if not messages:
        return messages

    processed = []
    for msg in messages:
        role = msg.get("role", "user")
        content = msg.get("content") or ""

        # Convert 'tool' or 'function' role to 'user' (tool responses are external input like user messages)
        # Note: 'function' is the older OpenAI format, 'tool' is the newer format
        if role in ("tool", "function"):
            role = "user"
            content = f"[Tool Result] {content}"

        # Check if we need to merge with previous message (same role after conversion)
        if processed and processed[-1]["role"] == role:
            # Merge consecutive same-role messages
            processed[-1]["content"] = f"{processed[-1]['content']}\n\n{content}"
        else:
            processed.append({"role": role, "content": content})

    return processed


def fix_message_alternation(messages: List[Dict[str, str]]) -> List[Dict[str, str]]:
    """Fix messages to ensure proper user/assistant alternation.

    This handles:
    1. Consecutive same-role messages (merge them)
    2. System → assistant without user in between (insert placeholder user)
    3. Assistant at start without preceding user (insert placeholder user)

    Args:
        messages: List of message dicts with 'role' and 'content' keys

    Returns:
        Fixed messages with proper alternation
    """
    if not messages:
        return messages

    # Step 1: Merge consecutive same-role messages
    merged = []
    for msg in messages:
        role = msg.get("role", "user")
        content = msg.get("content") or ""

        if merged and merged[-1]["role"] == role:
            # Merge content with newline separator
            merged[-1]["content"] = f"{merged[-1]['content']}\n{content}"
        else:
            merged.append({"role": role, "content": content})

    # Step 2: Handle alternation issues
    result = []
    for msg in merged:
        role = msg["role"]

        if role == "system":
            result.append(msg)
        elif role == "assistant":
            # Check if we need to insert a user message before assistant
            if not result or result[-1]["role"] in ("system", "assistant"):
                # Assistant without preceding user - insert placeholder
                result.append({"role": "user", "content": "[Continued]"})
            result.append(msg)
        elif role == "user":
            # Check for consecutive users (shouldn't happen after merge, but safety)
            if result and result[-1]["role"] == "user":
                result[-1]["content"] = f"{result[-1]['content']}\n{msg['content']}"
            else:
                result.append(msg)
        else:
            # Unknown role - just append
            result.append(msg)

    return result


def safe_apply_chat_template(
    tokenizer: AutoTokenizer,
    messages: List[Dict[str, Any]],
    tokenize: bool = False,
    add_generation_prompt: bool = False,
    **kwargs,
) -> str:
    """Safely apply chat template with automatic tool handling and alternation fixing.

    This function automatically:
    1. Detects if the tokenizer supports the 'tool_calls' field and serializes if needed
    2. Detects if the tokenizer supports the 'tool' role and converts if needed
    3. Fixes alternation issues (consecutive same-role, missing user before assistant)

    Use this instead of directly calling tokenizer.apply_chat_template() when
    messages may have tool_calls, tool role, or alternation issues.

    Args:
        tokenizer: The tokenizer to use
        messages: List of message dicts with 'role', 'content', and optionally 'tool_calls' keys
        tokenize: Whether to return token IDs (default False for string output)
        add_generation_prompt: Whether to add generation prompt
        **kwargs: Additional arguments passed to apply_chat_template

    Returns:
        Formatted conversation string (or token IDs if tokenize=True)

    Example:
        >>> messages = [
        ...     {"role": "user", "content": "What's 2+2?"},
        ...     {"role": "assistant", "content": "Let me calculate", "tool_calls": [...]},
        ...     {"role": "tool", "content": "4"},
        ...     {"role": "assistant", "content": "The answer is 4"}
        ... ]
        >>> # For Gemma (no tool support): auto-converts tool_calls to JSON in content
        >>> # For Llama 3.1+ (has tool support): uses native format
        >>> result = safe_apply_chat_template(tokenizer, messages)
    """
    # Check if messages have tool_calls field
    has_tool_calls = any(msg.get("tool_calls") for msg in messages)

    # Only serialize tool_calls if present AND tokenizer doesn't support them natively
    if has_tool_calls and not check_tool_calls_support(tokenizer):
        from autotrain import logger

        logger.debug("Tokenizer doesn't support 'tool_calls' field, serializing to content as JSON")
        messages = serialize_tool_calls_to_content(messages)

    # Check if messages have tool role
    has_tool = any(msg.get("role") == "tool" for msg in messages)

    # Only preprocess if we have tool messages AND tokenizer doesn't support them
    if has_tool and not check_tool_role_support(tokenizer):
        from autotrain import logger

        logger.debug("Tokenizer doesn't support 'tool' role, converting to 'user' with [Tool Result] prefix")
        messages = preprocess_messages_for_tool_role(messages)

    # Try to apply template
    try:
        return tokenizer.apply_chat_template(
            messages,
            tokenize=tokenize,
            add_generation_prompt=add_generation_prompt,
            **kwargs,
        )
    except Exception as e:
        error_str = str(e).lower()

        # Check if it's an alternation error - if so, try to fix it
        if "alternate" in error_str or "must be" in error_str:
            from autotrain import logger

            logger.debug(f"Alternation error detected: {e}, attempting to fix message structure")
            messages = fix_message_alternation(messages)

            try:
                return tokenizer.apply_chat_template(
                    messages,
                    tokenize=tokenize,
                    add_generation_prompt=add_generation_prompt,
                    **kwargs,
                )
            except Exception as e2:
                from autotrain import logger

                logger.warning(f"Failed to apply chat template even after fixing alternation: {e2}")
                raise e2

        # Different error, re-raise
        raise


def build_generation_prompt(
    messages: Union[List[Dict[str, str]], Conversation],
    format: Union[ChatFormat, str] = ChatFormat.CHATML,
    tokenizer: Optional[AutoTokenizer] = None,
    add_generation_prompt: bool = True,
) -> str:
    """
    Build a generation prompt from messages.

    Args:
        messages: List of message dicts or Conversation object
        format: Chat format to use
        tokenizer: Optional tokenizer for format detection
        add_generation_prompt: Whether to add generation prompt

    Returns:
        Formatted prompt string
    """
    # Convert to Conversation if needed, preserving tool_calls
    if isinstance(messages, list):
        conversation = Conversation(
            messages=[
                Message(
                    role=m["role"],
                    content=m.get("content") or "",
                    tool_calls=m.get("tool_calls"),
                    tool_call_id=m.get("tool_call_id"),
                )
                for m in messages
            ]
        )
    else:
        conversation = messages

    # Get appropriate renderer
    config = RenderConfig(
        format=format if isinstance(format, ChatFormat) else ChatFormat(format),
        add_generation_prompt=add_generation_prompt,
    )

    # Create dummy tokenizer if not provided
    if tokenizer is None:
        from transformers import AutoTokenizer

        tokenizer = AutoTokenizer.from_pretrained("gpt2")

    renderer = get_renderer(format, tokenizer, config)

    if add_generation_prompt:
        return renderer.build_generation_prompt(conversation)
    else:
        return renderer.render_conversation(conversation)


def parse_response(
    response: str,
    format: Union[ChatFormat, str] = ChatFormat.CHATML,
    tokenizer: Optional[AutoTokenizer] = None,
) -> str:
    """
    Parse a model response to extract content.

    Args:
        response: Raw model output
        format: Chat format used
        tokenizer: Optional tokenizer

    Returns:
        Parsed content string
    """
    if tokenizer is None:
        from transformers import AutoTokenizer

        tokenizer = AutoTokenizer.from_pretrained("gpt2")

    config = RenderConfig(format=format if isinstance(format, ChatFormat) else ChatFormat(format))
    renderer = get_renderer(format, tokenizer, config)

    return renderer.parse_response(response)


def get_stop_sequences(
    format: Union[ChatFormat, str] = ChatFormat.CHATML,
    tokenizer: Optional[AutoTokenizer] = None,
) -> List[str]:
    """
    Get stop sequences for a chat format.

    Args:
        format: Chat format
        tokenizer: Optional tokenizer

    Returns:
        List of stop sequences
    """
    if tokenizer is None:
        from transformers import AutoTokenizer

        tokenizer = AutoTokenizer.from_pretrained("gpt2")

    config = RenderConfig(format=format if isinstance(format, ChatFormat) else ChatFormat(format))
    renderer = get_renderer(format, tokenizer, config)

    return renderer.get_stop_sequences()


def build_supervised_example(
    messages: Union[List[Dict[str, str]], Conversation],
    tokenizer: AutoTokenizer,
    format: Union[ChatFormat, str] = ChatFormat.CHATML,
    max_length: int = 512,
    mask_system: bool = False,
    mask_user: bool = False,
    only_assistant: bool = True,
) -> Dict[str, torch.Tensor]:
    """
    Build a supervised training example with proper masking.

    Args:
        messages: Conversation messages
        tokenizer: Tokenizer to use
        format: Chat format
        max_length: Maximum sequence length
        mask_system: Whether to mask system messages
        mask_user: Whether to mask user messages
        only_assistant: Whether to only train on assistant messages

    Returns:
        Dictionary with input_ids, attention_mask, labels, and token_weights
    """
    # Convert to Conversation if needed, preserving tool_calls
    if isinstance(messages, list):
        conversation = Conversation(
            messages=[
                Message(
                    role=m["role"],
                    content=m.get("content") or "",
                    tool_calls=m.get("tool_calls"),
                    tool_call_id=m.get("tool_call_id"),
                )
                for m in messages
            ]
        )
    else:
        conversation = messages

    # Configure rendering
    config = RenderConfig(
        format=format if isinstance(format, ChatFormat) else ChatFormat(format),
        max_length=max_length,
        mask_system=mask_system,
        mask_user=mask_user,
        only_assistant=only_assistant,
    )

    # Get renderer and build example
    renderer = get_renderer(format, tokenizer, config)
    return renderer.build_supervised_example(conversation, tokenizer)


def apply_token_weights(
    input_ids: torch.Tensor,
    weights: Union[List[TokenWeight], torch.Tensor],
    default_weight: float = 1.0,
) -> torch.Tensor:
    """
    Apply token-level weights to input IDs.

    Args:
        input_ids: Input token IDs
        weights: List of TokenWeight objects or weight tensor
        default_weight: Default weight for unspecified tokens

    Returns:
        Weight tensor matching input_ids shape
    """
    if isinstance(weights, torch.Tensor):
        return weights

    # Initialize weight tensor
    weight_tensor = torch.full_like(input_ids, default_weight, dtype=torch.float)

    # Apply individual weights
    for weight in weights:
        weight_tensor = weight.apply_to_tensor(weight_tensor)

    return weight_tensor


def detect_chat_format(
    model_name: str,
    tokenizer: Optional[AutoTokenizer] = None,
) -> ChatFormat:
    """
    Auto-detect chat format from model name or tokenizer.

    Args:
        model_name: Name of the model
        tokenizer: Optional tokenizer with chat template

    Returns:
        Detected ChatFormat
    """
    model_lower = model_name.lower()

    # Check model name patterns
    if "chatml" in model_lower or "qwen" in model_lower:
        return ChatFormat.CHATML
    elif "alpaca" in model_lower:
        return ChatFormat.ALPACA
    elif "vicuna" in model_lower:
        return ChatFormat.VICUNA
    elif "zephyr" in model_lower:
        return ChatFormat.ZEPHYR
    elif "llama" in model_lower:
        return ChatFormat.LLAMA
    elif "mistral" in model_lower:
        return ChatFormat.MISTRAL

    # Check tokenizer chat template if available
    if tokenizer and hasattr(tokenizer, "chat_template"):
        template = str(tokenizer.chat_template)
        if "im_start" in template:
            return ChatFormat.CHATML
        elif "INST" in template:
            return ChatFormat.LLAMA
        elif "### Instruction" in template:
            return ChatFormat.ALPACA

    # Default to ChatML
    return ChatFormat.CHATML


def convert_dataset_to_conversations(
    dataset: Any,
    text_column: str = "text",
    conversation_column: Optional[str] = "conversations",
) -> List[Conversation]:
    """
    Convert a dataset to list of Conversation objects.

    Args:
        dataset: HuggingFace dataset or list of examples
        text_column: Column containing text
        conversation_column: Column containing conversation data

    Returns:
        List of Conversation objects
    """
    conversations = []

    # Handle different dataset formats
    if hasattr(dataset, "column_names"):
        # HuggingFace dataset
        if conversation_column in dataset.column_names:
            # Already has conversations
            for example in dataset:
                conv_data = example[conversation_column]
                if isinstance(conv_data, str):
                    import json

                    conv_data = json.loads(conv_data)
                conversations.append(Conversation.from_dict(conv_data))
        elif text_column in dataset.column_names:
            # Plain text - convert to single turn
            for example in dataset:
                text = example[text_column]
                conv = Conversation(
                    messages=[Message(role="user", content=""), Message(role="assistant", content=text)]
                )
                conversations.append(conv)
    elif isinstance(dataset, list):
        # List of examples
        for example in dataset:
            if isinstance(example, dict):
                if "messages" in example:
                    conversations.append(Conversation.from_dict(example))
                elif "text" in example:
                    conv = Conversation(
                        messages=[Message(role="user", content=""), Message(role="assistant", content=example["text"])]
                    )
                    conversations.append(conv)
            elif isinstance(example, str):
                conv = Conversation(
                    messages=[Message(role="user", content=""), Message(role="assistant", content=example)]
                )
                conversations.append(conv)

    return conversations


def create_chat_template(
    format: Union[ChatFormat, str] = ChatFormat.CHATML,
    system_template: Optional[str] = None,
    user_template: Optional[str] = None,
    assistant_template: Optional[str] = None,
) -> str:
    """
    Create a Jinja2 chat template for a specific format.

    Args:
        format: Chat format to use
        system_template: Custom system template
        user_template: Custom user template
        assistant_template: Custom assistant template

    Returns:
        Jinja2 template string
    """
    format = format if isinstance(format, ChatFormat) else ChatFormat(format)

    if format == ChatFormat.CHATML:
        return """{% for message in messages %}
{{'<|im_start|>' + message['role'] + '\n' + message['content'] + '<|im_end|>' + '\n'}}
{% endfor %}
{% if add_generation_prompt %}
{{ '<|im_start|>assistant\n' }}
{% endif %}"""

    elif format == ChatFormat.ALPACA:
        return """{% if messages[0]['role'] == 'system' %}
{{ messages[0]['content'] + '\n\n' }}
{% endif %}
{% for message in messages %}
{% if message['role'] == 'user' %}
{{ '### Instruction:\n' + message['content'] + '\n\n' }}
{% elif message['role'] == 'assistant' %}
{{ '### Response:\n' + message['content'] + '\n\n' }}
{% endif %}
{% endfor %}
{% if add_generation_prompt %}
{{ '### Response:\n' }}
{% endif %}"""

    elif format == ChatFormat.LLAMA:
        return """{% if messages[0]['role'] == 'system' %}
{{ '<s>[INST] <<SYS>>\n' + messages[0]['content'] + '\n<</SYS>>\n\n' }}
{% else %}
{{ '<s>[INST] ' }}
{% endif %}
{% for message in messages %}
{% if message['role'] == 'user' %}
{{ message['content'] + ' [/INST]' }}
{% elif message['role'] == 'assistant' %}
{{ ' ' + message['content'] + ' </s>' }}
{% endif %}
{% endfor %}
{% if add_generation_prompt %}
{{ ' ' }}
{% endif %}"""

    elif format == ChatFormat.CUSTOM:
        # Build custom template from provided templates
        system = system_template or "{content}"
        user = user_template or "User: {content}"
        assistant = assistant_template or "Assistant: {content}"

        # Build Jinja2 template with custom formats
        # Use Python string formatting to inject the template patterns
        template = "{% for message in messages %}\n"
        template += "{% if message['role'] == 'system' %}\n"
        template += "{{ '" + system.replace("'", "\\'") + "'.replace('{content}', message['content']) }}\n"
        template += "{% elif message['role'] == 'user' %}\n"
        template += "{{ '" + user.replace("'", "\\'") + "'.replace('{content}', message['content']) }}\n"
        template += "{% elif message['role'] == 'assistant' %}\n"
        template += "{{ '" + assistant.replace("'", "\\'") + "'.replace('{content}', message['content']) }}\n"
        template += "{% endif %}\n"
        template += "{% endfor %}"

        return template

    else:
        # Default to a simple format
        return """{% for message in messages %}
{{ message['role'] + ': ' + message['content'] + '\n' }}
{% endfor %}"""
